def is_inside_notebook():
    try:
        from IPython import get_ipython

        return get_ipython() is not None
    except ModuleNotFoundError:
        return False


def is_inside_colab():
    if not is_inside_notebook():
        return False
    from IPython import get_ipython

    return "google.colab" in get_ipython().extension_manager.loaded



import enum
from abc import ABCMeta
from typing import Set, Optional


class ReservedTags(enum.Enum):
    """:meta private:"""

    ANNOTATION = "annotation"
    DOCUMENT = "text_document"
    THUMBNAIL_VIZ = "thumbnail_viz"
    VIDEO = "video"
    AUDIO = "audio"
    IMAGE = "image"
    PDF = "pdf"
    TEXT = "text"


# These are the base primitives that the data engine database is capable of storing
class MetadataFieldType(enum.Enum):
    """
    Backing types in the Data Engine's database
    """

    BOOLEAN = "BOOLEAN"
    """Python's ``bool``"""
    INTEGER = "INTEGER"
    """Python's ``int``"""
    DATETIME = "DATETIME"
    """Python's ``datetime.datetime``"""
    FLOAT = "FLOAT"
    """Python's ``float``"""
    STRING = "STRING"
    """Python's ``str``"""
    BLOB = "BLOB"
    """Python's ``bytes``"""


class ThumbnailType(enum.Enum):
    """
    Thumbnail types for visualization

    :meta private:
    """

    VIDEO = "video"
    AUDIO = "audio"
    IMAGE = "image"
    PDF = "pdf"
    TEXT = "text"


class DagshubDataType(metaclass=ABCMeta):
    """
    Inheritors of this ABC define custom types

    They are backed by a primitive type, but they also may have additional tags, which we use to enhance the experience.

    Attributes:
        backing_field_type: primitive type in the data engine database
        custom_tags: additional tags applied to this type
    """

    backing_field_type: Optional[MetadataFieldType] = None
    custom_tags: Optional[Set[str]] = None


class Int(DagshubDataType):
    """Basic python ``int``"""

    backing_field_type = MetadataFieldType.INTEGER


class DateTime(DagshubDataType):
    """Basic python ``datetime.datetime``

    .. note::
        Dagshub backend receives an integer millisecond timestamp (utc timestamp), and optionally a timezone.

        A metadata of type datetime is always stored in DB as a UTC time, when a query is done on this field
        there are 3 options:

        - Metadata was saved with a timezone, in which case it will be used.

        - Metadata was saved without a timezone, in which case UTC will be used.

        - with_time_zone specified a time zone and it will override whatever is in the database.

    Example::

        # dagshub client sends int(t.timestamp() * 1000) to the backend
        # and the +05:30 offset
        datapoints = datasource.all()
        t = dateutil.parser.parse("2022-04-05T15:30:00.99999+05:30")
        datapoints[path][name] = t
        datapoints[path].save()

        # or:
        # send only a millisecond timestamp, without timezone (will be saved as utc)
        datapoints[path][name] = int(dateutil.parser.parse("2022-04-05T15:30:00.99999+05:30").timestamp() * 1000)
    """

    backing_field_type = MetadataFieldType.DATETIME


class String(DagshubDataType):
    """Basic python ``str``"""

    backing_field_type = MetadataFieldType.STRING


class Blob(DagshubDataType):
    """
    Basic python ``bytes``

    .. note::
        DagsHub doesn't return the blob fields by default, instead returning their hashes.
        Check out :func:`Datapoint.get_blob() <dagshub.data_engine.model.datapoint.Datapoint.get_blob>`
        to learn how to download the blob value.
    """

    backing_field_type = MetadataFieldType.BLOB


class Float(DagshubDataType):
    """
    Basic python ``float``
    """

    backing_field_type = MetadataFieldType.FLOAT


class Bool(DagshubDataType):
    """
    Basic python ``bool``
    """

    backing_field_type = MetadataFieldType.BOOLEAN


class LabelStudioAnnotation(DagshubDataType):
    """
    LabelStudio annotation. Backing type is blob.
    Has the annotation tag set.

    Annotations of this type get automatically converted in the metadata into
    :class:`MetadataAnnotations <dagshub.data_engine.annotation.metadata.MetadataAnnotations>`
    objects that simplify adding and saving annotations.
    """

    backing_field_type = MetadataFieldType.BLOB
    custom_tags = {ReservedTags.ANNOTATION.value}


class Voxel51Annotation(DagshubDataType):
    """
    Voxel51 annotation. Backing type is blob.
    Has the annotation tag set.
    """

    backing_field_type = MetadataFieldType.BLOB
    custom_tags = {ReservedTags.ANNOTATION.value}


class Document(DagshubDataType):
    """
    Field with large text values that is stored as a blob.
    Document fields can't be filtered on,
    but allow you to store arbitrarily large text longer than allowed 512 characters
    """

    backing_field_type = MetadataFieldType.BLOB
    custom_tags = {ReservedTags.DOCUMENT.value}



from dataclasses import dataclass
import functools
from typing import Any, Dict, List
from dagshub.data_engine.client.query_builder import GqlQuery, ParamValidator


class GqlIntrospections:
    @staticmethod
    @functools.lru_cache()
    def obj_fields() -> GqlQuery:
        q = (
            GqlQuery()
            .operation("query", name="introspection")
            .query("__schema")
            .fields(
                [
                    GqlQuery()
                    .fields(
                        name="types",
                        fields=[
                            "name",
                            GqlQuery()
                            .fields(
                                name="inputFields",
                                fields=["name", GqlQuery().fields(name="type", fields=["name"]).generate()],
                            )
                            .generate(),
                            GqlQuery()
                            .fields(
                                name="fields",
                                fields=["name", GqlQuery().fields(name="type", fields=["name"]).generate()],
                            )
                            .generate(),
                        ],
                    )
                    .generate()
                ]
            )
        )
        return q


@dataclass
class Field:
    name: str


@dataclass
class IntrospectionType:
    name: str
    inputFields: List[Field]
    fields: List[Field]

    def __hash__(self):
        return hash(self.hash_str())

    def hash_str(self):
        return self.name + "".join([f.name for f in self.fields]) + "".join([f.name for f in self.inputFields])


@dataclass
class TypesIntrospection:
    types: List[IntrospectionType]

    def __hash__(self):
        return hash("".join([t.hash_str() for t in self.types]))


class Validators:
    @staticmethod
    def query_input_validator(queryInputField: str):
        def validator(params: Dict[str, Any], query_input_introspection: TypesIntrospection):
            introspect_query_input_fields = Validators.get_fields(
                query_input_introspection, "QueryInput", "inputFields"
            )

            # Get sent fields
            query_input = params.get(queryInputField)
            if query_input is None:
                return
            sent_fields = query_input.keys()
            # Check serialized query input fields exist in introspection
            if not all([f in introspect_query_input_fields for f in sent_fields]):
                unsupported_fields = [f for f in sent_fields if f not in introspect_query_input_fields]
                raise ValueError(f"QueryInput fields are not supported: {unsupported_fields}")

        return validator

    @staticmethod
    def get_fields(
        datapoints_connection_introspection: TypesIntrospection, type_name: str, attr: str = "fields"
    ) -> List[str]:
        datapoints_connection_fields = [f for f in datapoints_connection_introspection.types if f.name == type_name]
        if len(datapoints_connection_fields) == 0:
            raise ValueError(f"{type_name} is not defined")
        datapoints_connection_fields = getattr(datapoints_connection_fields[0], attr)
        if datapoints_connection_fields is None:
            raise ValueError(f"{type_name} is not defined")
        return [f.name for f in datapoints_connection_fields]

    @staticmethod
    def filter_supported_fields(
        fields: List[str], response_obj_name: str, introspection: TypesIntrospection
    ) -> List[str]:
        root_fields = [f.split(" ")[0] for f in fields]
        supported_fields = Validators.get_fields(introspection, response_obj_name)
        return [fields[i] for i, rf in enumerate(root_fields) if rf in supported_fields]

    @staticmethod
    def has_type_validator(type_name: str) -> "ParamValidator":
        def validator(_: Dict[str, Any], introspection: TypesIntrospection):
            if type_name not in [t.name for t in introspection.types]:
                raise ValueError(
                    f"{type_name} is not available on the backend. Make sure your DagsHub instance is up to date"
                )

        return validator



import datetime
import functools
from typing import Optional, Any, Dict, Union, List

from dagshub.data_engine.client.gql_introspections import Validators, TypesIntrospection
from dagshub.data_engine.client.query_builder import GqlQuery


class GqlQueries:
    @staticmethod
    @functools.lru_cache()
    def datasource() -> GqlQuery:
        q = (
            GqlQuery()
            .operation("query", name="datasource", input={"$id": "ID", "$name": "String"})
            .query(
                "datasource",
                input={
                    "id": "$id",
                    "name": "$name",
                },
            )
            .fields(
                [
                    "id",
                    "name",
                    "rootUrl",
                    "integrationStatus",
                    "preprocessingStatus",
                    "metadataFields {name valueType multiple tags}" "type",
                ]
            )
        )
        return q

    @staticmethod
    def datasource_params(id: Optional[Union[int, str]], name: Optional[str]) -> Dict[str, Any]:
        return {
            "id": id,
            "name": name,
        }

    @staticmethod
    @functools.lru_cache()
    def datasource_query(include_metadata: bool, introspection: "TypesIntrospection") -> GqlQuery:
        metadata_fields = ""
        if include_metadata:
            # Filter out the unavailable fields
            queryable_fields = ["key", "value", "timeZone"]
            queryable_fields = Validators.filter_supported_fields(queryable_fields, "MetadataField", introspection)
            metadata_fields = "metadata " + "{" + " ".join(queryable_fields) + "}"
        q = (
            GqlQuery()
            .operation(
                "query",
                name="datasourceQuery",
                input={
                    "$datasource": "ID!",
                    "$queryInput": "QueryInput",
                    "$first": "Int",
                    "$after": "String",
                },
            )
            .query(
                "datasourceQuery",
                input={
                    "datasource": "$datasource",
                    "filter": "$queryInput",
                    "first": "$first",
                    "after": "$after",
                },
            )
        )
        fields = [
            f"edges {{ node {{ id path {metadata_fields} }} }}",
            "pageInfo { hasNextPage endCursor } queryDataTime",
        ]
        if include_metadata:
            fields.append("selectFields { name originalName autoGenerated valueType asOf multiple tags }")
        q = q.fields(Validators.filter_supported_fields(fields, "DatapointsConnection", introspection))
        q = q.param_validator(Validators.query_input_validator("queryInput"))

        return q

    @staticmethod
    def datasource_query_params(
        datasource_id: Union[int, str], query_input: Dict[str, Any], first: Optional[int], after: Optional[str]
    ) -> Dict[str, Any]:
        return {
            "datasource": datasource_id,
            "queryInput": query_input,
            "first": first,
            "after": after,
        }

    @staticmethod
    @functools.lru_cache()
    def dataset() -> GqlQuery:
        q = (
            GqlQuery()
            .operation("query", name="dataset", input={"$id": "ID", "$name": "String"})
            .query(
                "dataset",
                input={
                    "dataset": "$id",
                    "name": "$name",
                },
            )
            .fields(
                [
                    "id",
                    "name",
                    "datasource {id name rootUrl integrationStatus preprocessingStatus "
                    "metadataFields {name valueType multiple tags} type}",
                    "datasetQuery",
                ]
            )
        )
        return q

    @staticmethod
    def dataset_params(id: Optional[Union[int, str]], name: Optional[str]) -> Dict[str, Any]:
        return {
            "id": id,
            "name": name,
        }

    @staticmethod
    @functools.lru_cache()
    def datapoint_history() -> GqlQuery:
        # TODO: introspect and check if the query is available
        q = (
            GqlQuery()
            .operation(
                "query",
                name="datapointHistory",
                input={
                    "$datasource": "ID!",
                    "$opts": "DatapointHistoryInput!",
                },
            )
            .query(
                "datapointHistory",
                input={
                    "datasource": "$datasource",
                    "opts": "$opts",
                },
            )
            .fields(
                [
                    "edges { node { path history { timestamp } } }",
                    "pageInfo { hasNextPage endCursor }",
                ]
            )
        )
        q.param_validator(Validators.has_type_validator("DatapointHistory"))
        return q

    @staticmethod
    def datapoint_history_params(
        datasource_id: Union[int, str],
        datapoints: List[str],
        fields: Optional[List[str]],
        from_time: Optional[datetime.datetime],
        to_time: Optional[datetime.datetime],
        after: Optional[str],
        first: Optional[int],
    ) -> Dict[str, Any]:
        return {
            "datasource": datasource_id,
            "opts": {
                "datapoints": datapoints,
                "fields": fields,
                "fromTime": int(from_time.timestamp()) if from_time else None,
                "toTime": int(to_time.timestamp()) if to_time else None,
                "after": after if after else None,
                "first": first if first else None,
            },
        }



import logging
import os.path
import signal
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import partial
from pathlib import Path
from typing import Tuple, Callable, Optional, List, Union, Dict

from httpx import Auth, Response
from tenacity import stop_after_attempt, wait_exponential, before_sleep_log, retry, retry_if_exception

from dagshub.common import config

from typing import Literal

import re
import rich.progress

from dagshub.auth import get_authenticator
from dagshub.common.helpers import http_request
from dagshub.common.rich_util import get_rich_progress

logger = logging.getLogger(__name__)

DownloadFunctionType = Callable[[str, Path], None]

storage_download_url_regex = re.compile(
    r".*/api/v1/repos/(?P<user>[\w\-_.]+)/(?P<repo>[\w\-_.]+)/storage/raw/(?P<proto>s3|gs|azure)/"
    r"(?P<bucket>[a-z0-9.-]+)/(?P<path>.*)"
)


def enable_gcs_bucket_downloader(client=None):
    """
    Enables downloading storage items using the Google Cloud Storage client,\
    instead of going through DagsHub's server.

    For custom clients use :func:`add_bucket_downloader` function.

    Args:
        client: a `google.cloud.storage.Client \
        <https://cloud.google.com/storage/docs/reference/libraries#client-libraries-install-python>`_\
        from the ``google-cloud-storage`` package.\
        If client isn't specified, the default parameterless constructor is used
    """
    if client is None:
        from google.cloud import storage

        client = storage.Client()

    def get_fn(bucket_name, bucket_path) -> bytes:
        bucket = client.bucket(bucket_name)
        blob = bucket.blob(bucket_path).download_as_bytes()
        return blob

    add_bucket_downloader("gs", get_fn)


def enable_s3_bucket_downloader(client=None):
    """
    Enables downloading storage items using the AWS Boto3 client,\
    instead of going through DagsHub's server.

    For custom clients use :func:`add_bucket_downloader` function.

    Args:
        client: a `boto3.client \
        <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/boto3.html#boto3.client>`_.\
        If client isn't specified, the default parameterless constructor is used.
    """

    if client is None:
        import boto3

        client = boto3.client("s3")

    def get_fn(bucket, path) -> bytes:
        resp = client.get_object(Bucket=bucket, Key=path)
        return resp["Body"].read()

    add_bucket_downloader("s3", get_fn)


def enable_azure_container_downloader(account_url=None, client=None):
    """
    Enables downloading storage items using the Azure Blob Storage client,\
    instead of going through DagsHub's server.

    For custom clients use :func:`add_bucket_downloader` function.

    Args:
        account_url: an azure storage account url, of the form ``https://<storage-account-name>.blob.core.windows.net``
        client: preconfigured `azure.storage.blob.BlobServiceClient \
        <https://learn.microsoft.com/en-us/python/api/overview/azure/\
        storage-blob-readme?view=azure-python#create-the-client>`_.\
        If client isn't specified, the default parameterless constructor is used.\
        If specified, ``account_url`` is disregarded, and the client is used.
    """
    if account_url is None and client is None:
        raise TypeError("missing required argument 'account_url' or 'client'")

    import io

    if client is None:
        from azure.storage.blob import BlobServiceClient
        from azure.identity import DefaultAzureCredential

        client = BlobServiceClient(account_url, credential=DefaultAzureCredential())

    def get_fn(bucket, path) -> bytes:
        blob_client = client.get_blob_client(container=bucket, blob=path)
        stream = io.BytesIO()
        blob_client.download_blob().readinto(stream)
        return stream

    add_bucket_downloader("azure", get_fn)


def download_url_to_bucket_path(url: str) -> Optional[Tuple[str, str, str]]:
    """
    Gets a storage download URL of a dagshub file, returns a tuple of (protocol, bucket, path_in_bucket)
    """
    matches = storage_download_url_regex.match(url)
    if matches is None:
        return None
    groups = matches.groupdict()
    return groups["proto"], groups["bucket"], groups["path"]


class DownloadError(Exception):
    def __init__(self, response: Response):
        self.response = response
        super().__init__(f"Download failed with status code {response.status_code}")


def is_download_server_error(error: BaseException) -> bool:
    if not isinstance(error, DownloadError):
        return False
    return error.response.status_code >= 500


@retry(
    retry=retry_if_exception(is_download_server_error),
    stop=stop_after_attempt(5),
    wait=wait_exponential(multiplier=1, min=4, max=10),
    before_sleep=before_sleep_log(logger, logging.WARNING),
)
def _dagshub_download(url: str, auth: Auth) -> bytes:
    resp = http_request("GET", url, auth=auth, timeout=600)
    if resp.status_code == 200:
        return resp.content
    raise DownloadError(resp)


BucketDownloaderFuncType = Callable[[str, str], bytes]

_bucket_downloader_map: Dict[str, BucketDownloaderFuncType] = {}
_default_downloader: Optional[Callable[[str], bytes]] = None


def add_bucket_downloader(proto: Literal["gs", "s3", "azure"], func: BucketDownloaderFuncType):
    """
    Add your own custom connected bucket downloader.

    Args:
        proto: Protocol for which you're adding the downloader.\
            This function will handle **all** download requests to this protocol.
        func: Function that receives the name of the bucket and the path to the object and returns the object content\
            in ``bytes``.

    .. warning::
        The ``func`` function will be used in a ThreadPool, so it needs to be picklable.
    """
    if proto in _bucket_downloader_map:
        logger.warning(f"Protocol {proto} already has a custom downloader function specified, overwriting it")
    _bucket_downloader_map[proto] = func


def _download_wrapper(url: str, location: Path, skip_if_exists: bool):
    if skip_if_exists and os.path.exists(location):
        return

    # Download the file
    # Check if it's a bucket
    bucket_tuple = download_url_to_bucket_path(url)
    assert _default_downloader is not None
    if bucket_tuple is None:
        # Not a bucket path - download regularly
        content = _default_downloader(url)
    else:
        # Bucket path - try to look if there's a custom downloader
        proto, bucket_name, bucket_path = bucket_tuple
        bucket_downloader = _bucket_downloader_map.get(proto)
        if bucket_downloader is None:
            content = _default_downloader(url)
        else:
            content = bucket_downloader(bucket_name, bucket_path)

    location.parent.mkdir(parents=True, exist_ok=True)
    with open(location, "wb") as f:
        f.write(content)


def _ensure_default_downloader_exists():
    """
    Checks that the default dagshub download function exists and prepares it otherwise
    """
    global _default_downloader
    if _default_downloader is None:
        auth = get_authenticator()
        _default_downloader = partial(_dagshub_download, auth=auth)


def download_files(
    files: List[Tuple[str, Union[str, Path]]],
    download_fn: Optional[DownloadFunctionType] = None,
    threads=config.download_threads,
    skip_if_exists=True,
):
    """
    Download files using multithreading

    Parameters:
        files: list of (download_url: str, file_location: str or Path)
        download_fn: Optional function that will download the file. Needs to receive the two arguments:
            download url and Path where to save the file
            If function is not specified, then a default function that downloads a file with DagsHub credentials is used
            CAUTION: function needs to be pickleable since we're using ThreadPool to execute
        threads: number of threads to run this function on, defaults to the config value of download_threads (32)
        skip_if_exists: skip the download if the file exists (only for the default downloader)
    """
    _ensure_default_downloader_exists()

    # Convert string paths to Path objects
    for i, file_tuple in enumerate(files):
        if isinstance(file_tuple[1], str):
            files[i] = (file_tuple[0], Path(file_tuple[1]))

    if download_fn is None:
        download_fn = partial(_download_wrapper, skip_if_exists=skip_if_exists)

    if len(files) > 1:
        # Multiple files - multithreaded download
        progress = get_rich_progress(rich.progress.MofNCompleteColumn(), transient=False)
        task = progress.add_task("Downloading files...", total=len(files))

        with progress:
            with ThreadPoolExecutor(max_workers=threads) as tp:

                def cancel_download(*args):
                    logger.warning("Interrupt received - shutting down downloader")
                    tp.shutdown(wait=False, cancel_futures=True)

                orig_interrupt = None
                try:
                    orig_interrupt = signal.signal(signal.SIGINT, cancel_download)
                # ValueError means the function is not running from the main thread.
                # TODO: figure out a workaround
                except ValueError:
                    pass

                futures = [tp.submit(download_fn, url, location) for (url, location) in files]
                for f in as_completed(futures):
                    exc = f.exception()
                    if exc is not None:
                        logger.warning(f"Got exception {type(exc)} while downloading file: {exc}")
                    progress.update(task, advance=1)

                if orig_interrupt is not None:
                    signal.signal(signal.SIGINT, orig_interrupt)

    elif len(files) == 1:
        # Single file - don't bother with the multithreading, just download the file
        url, location = files[0]
        try:
            download_fn(url, location)
        except Exception as exc:
            logger.warning(f"Got exception {type(exc)} while downloading file: {exc}")



import fnmatch
import json
import logging
import os
import posixpath
import time
import urllib
from http import HTTPStatus
from pathlib import Path, PurePosixPath, PurePath
from typing import Union, Tuple, BinaryIO, Dict, Optional, Any, List, cast
from concurrent.futures import ThreadPoolExecutor
from urllib.parse import urljoin

import httpx
import rich.progress
import rich.status
from tenacity import retry, retry_if_exception_type, wait_fixed, stop_after_attempt

import dagshub.auth
from dagshub.auth.token_auth import EnvVarDagshubToken
from dagshub.common import config, rich_console
from dagshub.common.api.repo import RepoAPI, BranchNotFoundError
from dagshub.common.helpers import log_message
from dagshub.upload.errors import determine_upload_api_error, InternalServerErrorError
from dagshub.common.rich_util import get_rich_progress
from dagshub.repo_bucket import get_repo_bucket_client

# todo: handle api urls in common package
CONTENT_UPLOAD_URL = "api/v1/repos/{owner}/{reponame}/content/{branch}/{path}"
FILES_UI_URL = "{owner}/{reponame}/src/{branch}/{path}"
DEFAULT_COMMIT_MESSAGE = "Upload files using DagsHub client"
DEFAULT_DATASET_COMMIT_MESSAGE = "Initial dataset commit"
REPO_CREATE_URL = "api/v1/user/repos"
ORG_REPO_CREATE_URL = "api/v1/org/{orgname}/repos"
USER_INFO_URL = "api/v1/user"
DEFAULT_DATA_DIR_NAME = "data"
logger = logging.getLogger(__name__)

s = httpx.Client()
s.timeout = config.http_timeout
s.follow_redirects = True
s.headers.update(config.requests_headers)

FileUploadStruct = Tuple[str, BinaryIO]


def create_dataset(repo_name: str, local_path: str, glob_exclude: str = "", org_name: str = "", private=False):
    """
    Create a new repository on DagsHub and upload an entire folder dataset to it

    Args:
        repo_name: Name of the repository to be created.
        local_path: local path where the dataset to upload is located.
        glob_exclude: glob pattern to exclude certain files from being uploaded.
        org_name (optional): Organization that will own the repo. Alternative to creating a repository owned by you.
        private: Set to ``True`` to make the repository private.

    Returns:
        :class:`.Repo`: Repo object of the repository created.
    """
    repo = create_repo(repo_name, org_name=org_name, private=private)
    dir = repo.directory(repo_name)
    dir.add_dir(local_path, glob_exclude, commit_message=DEFAULT_DATASET_COMMIT_MESSAGE)
    return repo


def add_dataset_to_repo(repo: "Repo", local_path: str, data_dir: str = DEFAULT_DATA_DIR_NAME):
    """
    Given a repository created on dagshub - upload a dataset folder into it

    Args:
        repo: repository created beforehand
        local_path: local path where the dataset to upload is located
        data_dir: name of data directory that will be created inside repo
    """
    dir = repo.directory(data_dir)
    dir.add_dir(local_path, commit_message=DEFAULT_DATASET_COMMIT_MESSAGE)


def create_repo(
    repo_name: str,
    org_name: str = "",
    description: str = "",
    private: bool = False,
    auto_init: bool = False,
    gitignores: str = "Python",
    license: str = "",
    readme: str = "",
    template: str = "custom",
    host: str = "",
):
    """
    Creates a repository on DagsHub for the current user or an organization passed as an argument

    Args:
        repo_name: Name of the repository to be created.
        org_name (optional): Organization that will own the repo. Alternative to creating a repository owned by you.
        description: Repository description.
        private: Set to ``True`` to make repository private.
        auto_init: Set to True to create an initial commit with README, .gitignore and LICENSE.
        gitignores: Which gitignore template(s) to use in a comma separated string.
        license: Which license file to use.
        readme: Readme template to initialize with.
        template: Which project template to use, options are:

            - ``"none"`` - creates an empty repo
            - ``"custom"`` - creates a repo with your specified ``gitignores``, ``license`` and ``readme``
            - ``"notebook-template"``
            - ``"cookiecutter-mlops"``
            - ``"cookiecutter-dagshub-dvc"``

            By default, creates an empty repo if none of ``gitignores``, ``license`` or ``readme`` were provided.
            Otherwise, the template is ``"custom"``.

        host: URL of the DagsHub instance to host the repo on.

    .. note::
        To learn more about the templates, visit https://dagshub.com/docs/feature_guide/project_templates/

    Returns:
        :class:`.Repo`: Repo object of the repository created.
    """
    if template == "":
        template = "none"

    host = host or config.host

    username = config.username
    password = config.password
    token = None
    if username is not None and password is not None:
        auth = username, password
    else:
        auth = dagshub.auth.get_authenticator()
        token = auth.token_text

    if auth is None:
        raise RuntimeError("You can't create a repository without being authenticated.")

    if (license != "" or readme != "" or gitignores != "") and template == "none":
        template = "custom"

    data = {
        "name": repo_name,
        "description": description,
        "private": private,
        "auto_init": auto_init,
        "gitignores": gitignores,
        "license": license,
        "readme": readme,
        "project_template": template,
    }

    url = REPO_CREATE_URL
    if org_name and not org_name.isspace():
        url = ORG_REPO_CREATE_URL.format(
            orgname=org_name,
        )

    res = s.post(urljoin(host, url), data=data, auth=auth)

    if res.status_code != HTTPStatus.CREATED:
        logger.error(f"Response ({res.status_code}):\n" f"{res.content.decode()}")
        if res.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
            raise RuntimeError("Repository name is invalid or it already exists.")
        else:
            raise RuntimeError("Failed to create the desired repository.")

    repo = res.json()
    return Repo(
        owner=repo["owner"]["login"],
        name=repo["name"],
        username=username,
        password=password,
        token=token,
        branch="main",
    )


def validate_owner_repo(owner_repo: str) -> Tuple[str, str]:
    parts = owner_repo.split("/")
    if len(parts) != 2:
        raise ValueError("repo needs to be in the format <repo-owner>/<repo-name>")
    return parts[0], parts[1]


def upload_files(
    repo: str,
    local_path: Union[str, os.PathLike],
    commit_message=DEFAULT_COMMIT_MESSAGE,
    remote_path: Optional[str] = None,
    bucket: bool = False,
    **kwargs,
):
    """
    Upload file(s) into a repository.

    Args:
        repo: Repo name in the form of ``<username>/<reponame>``.
        local_path: File or directory to be uploaded.
        commit_message (optional): Specify a commit message.
        remote_path: Specify the path to upload the file to.\
        Defaults to the relative component of ``local_path`` to CWD.
        bucket: Upload the file(s) to the DagsHub Storage bucket

    For kwarg docs look at :func:`Repo.upload() <dagshub.upload.Repo.upload>`.
    """
    owner, repo = validate_owner_repo(repo)
    repo_obj = Repo(owner, repo)
    repo_obj.upload(local_path, commit_message=commit_message, remote_path=remote_path, bucket=bucket, **kwargs)


def upload_file_to_s3(
    s3_client,
    local_path: Union[PurePath, str],
    bucket_name: str,
    remote_path: Union[PurePath, str],
    progress=None,
    task=None,
):
    """
    Upload a single file to S3.

    :param s3_client: An instance of boto3 S3 client
    :param local_path: The local path to the file
    :param bucket_name: The S3 bucket name
    :param remote_path: The S3 path where the file will be uploaded
    :param progress: (rich.progress.Progress, optional) A Rich library Progress instance for visual progress tracking.
    If provided, `task_id` must also be given. Default is None.
    :param task: The task ID associated with this upload task in the Rich Progress instance.
    Required if `progress` is provided. Default is None.
    """
    s3_client.upload_file(str(local_path), bucket_name, str(remote_path))
    if progress is not None and task is not None:
        progress.update(task, advance=1)
    else:
        log_message(f"Uploaded {local_path} to s3://{bucket_name}/{remote_path}")


class Repo:
    def __init__(
        self,
        owner: str,
        name: str,
        username: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        branch: Optional[str] = None,
    ):
        """
        Class that can be used to upload files into a repository on DagsHub

        .. warning::
            This class is not thread safe.
            Uploading files in parallel can lead to unexpected outcomes

        Args:
            owner: user or org that owns the repository.
            name: name of the repository.
            token (optional): Token to use for authentication. If unset, uses the cached token or goes through OAuth.
            username: Username to log in with (alternative to token).
            password: Password to log in with (alternative to token).
            branch: Branch to upload files to.
        """
        self.owner = owner
        self.name = name
        self.host = config.host

        self.username = username or config.username
        self.password = password or config.password
        self.token = token or config.token
        self.branch = branch

        self._api = RepoAPI(f"{owner}/{name}", host=self.host, auth=self.auth)

        # For mirror uploading: store the last revision for which we uploaded
        # When the last revision changes, that means the sync has been complete and we can upload a new batch
        self._last_upload_revision: Optional[str] = None
        self._last_upload_had_changes = False
        self._uploading_to_new_branch = False

        self.current_progress: Optional[rich.progress.Progress] = None

        if self.branch is None:
            logger.debug("Branch wasn't provided. Fetching default branch...")
            self.branch = self._api.default_branch
        logger.debug(f"Set branch: {self.branch}")

    def upload(
        self,
        local_path: Union[str, os.PathLike],
        commit_message=DEFAULT_COMMIT_MESSAGE,
        remote_path: Optional[str] = None,
        bucket: bool = False,
        **kwargs,
    ):
        """
        Upload a file or a directory to the repo.

        Args:
            local_path: Path to file or directory to be uploaded
            commit_message: Specify a commit message
            remote_path: Specify the path to upload the file/dir to.
                If unspecified, sets the value to the relative component of ``local_path`` to CWD.
                If ``local_path`` is not relative to CWD, ``remote_path`` is the last component of the ``local_path``
            bucket: Upload to the DagsHub Storage bucket (s3-compatible) without versioning, if this is set to true,
            commit_message will be ignored.

        The kwargs are the parameters of :func:`upload_files`
        """
        if commit_message is None:
            commit_message = DEFAULT_COMMIT_MESSAGE
        local_path = cast(Path, Path(local_path).resolve())

        if not remote_path:
            try:
                remote_path = str(local_path.relative_to(Path.cwd().resolve()))
                if remote_path == ".":
                    remote_path = local_path.name
            except ValueError:
                # local_path is outside cwd, use only its basename
                remote_path = local_path.name
        remote_path = cast(str, Path(remote_path).as_posix())

        if bucket:
            self.upload_files_to_bucket(local_path, remote_path, **kwargs)
        else:
            if local_path.is_dir():
                dir_to_upload = self.directory(remote_path)
                dir_to_upload.add_dir(str(local_path), commit_message=commit_message, **kwargs)
            else:
                file_to_upload = DataSet.get_file(str(local_path), PurePosixPath(remote_path))
                self.upload_files([file_to_upload], commit_message=commit_message, **kwargs)

    @retry(retry=retry_if_exception_type(InternalServerErrorError), wait=wait_fixed(3), stop=stop_after_attempt(5))
    def upload_files(
        self,
        files: List[FileUploadStruct],
        directory_path: str = "",
        commit_message: Optional[str] = DEFAULT_COMMIT_MESSAGE,
        versioning: str = "auto",
        new_branch: Optional[str] = None,
        last_commit: Optional[str] = None,
        force: bool = False,
        quiet: bool = False,
    ):
        """
        Upload a list of binary files to the specified directory.
        This function is lower level than :func:`upload`,
        but useful when you don't have the files stored on the filesystem.

        Args:
            files: List of Tuples of (path in repo, binaryIO) of files to upload
            directory_path: Directory in repo relative to which to upload files
            commit_message: Commit message
            versioning: Which versioning system to use to upload a file.
                Possible options: ``"git"``, ``"dvc"``, ``"auto"`` (default, best effort guess)
            new_branch: Create a new branch with this name
            last_commit: Consistency argument - last revision of the files you want to commit on top of.
                Exists to prevent accidental overwrites of data.
            force (bool): Force the upload of a file even if it is already present on the server.
                Sets last_commit to be the tip of the branch
            quiet (bool): Don't show messages about starting/successfully completing an upload.
                Set to True when uploading a directory
        """

        if commit_message is None:
            commit_message = DEFAULT_COMMIT_MESSAGE

        # Truncate the commit message because the max we allow is 100 symbols
        commit_message = commit_message[:100]

        data = {
            "commit_choice": "direct",
            "commit_summary": commit_message,
            "versioning": versioning,
            "last_commit": last_commit,
            "is_dvc_dir": directory_path != "" and versioning != "git",
        }

        if self._api.is_mirror and new_branch is not None:
            # If not uploading to a new branch, and we're in a mirror - wait for the sync to complete
            self._poll_mirror_up_to_date()

        # Unset the new branch upload flag if it was set
        # (do that only after mirror poll so commit 2 on a new branch pushes correctly)
        if self._uploading_to_new_branch:
            self._uploading_to_new_branch = False

        upload_url = self.get_request_url(directory_path)

        try:
            self._last_upload_revision = self._api.last_commit_sha(self.branch)
        # New branch does not have a commit yet, so "last_upload_revision" will be None initially
        # NOTE: checking for new_branch is not enough, because it doesn't take into account
        # uploading to completely blank repos
        except BranchNotFoundError:
            self._uploading_to_new_branch = True
            log_message(
                f"Uploading to a new branch {self.branch}, "
                f"splitting it off from the default branch {self._api.default_branch}",
                logger,
            )
            upload_url = self.get_request_url(directory=directory_path, branch=self._api.default_branch)
            new_branch = self.branch

        if new_branch is not None:
            data.update(
                {
                    "commit_choice": "commit-to-new-branch",
                    "new_branch_name": new_branch,
                }
            )

        if force:
            data["last_commit"] = self._last_upload_revision

        if not quiet:
            log_message(f'Uploading files ({len(files)}) to "{self._api.full_name}"...', logger)
        res = s.put(
            upload_url,
            data=data,
            files=[("files", file) for file in files],
            auth=self.auth,
            timeout=None,
        )
        self._log_upload_details(data, res, files, quiet)

        # The ETag header contains the hash of the uploaded commit,
        # check against the one we have to determine if anything changed
        if "ETag" in res.headers:
            new_tip = res.headers["ETag"]
            self._last_upload_had_changes = new_tip != self._last_upload_revision

    def _log_upload_details(self, data: Dict[str, Any], res: httpx.Response, files, quiet: bool):
        """
        The _log_upload_details function debug logs the request URL, data, and files.
        It also prints for the user the status of their upload if it was successful
        If the response is 4xx/5xx it raises an error.

        Args:
            data: Executed request's body
            res: Server's response
            files: Uploaded file contents
            quiet: Log successful upload
        """

        logger.debug(
            f"Request URL: {res.request.url}\n"
            f"Data:\n{json.dumps(data, indent=4)}\n"
            f"Files:\n{json.dumps(list(map(str, files)), indent=4)}"
        )

        if res.status_code == HTTPStatus.OK:
            if "ETag" in res.headers:
                new_tip = res.headers["ETag"]
                if new_tip == self._last_upload_revision:
                    log_message("Upload successful, content was identical and no new commit was created", logger)
                    return
            if not quiet:
                log_message("Upload finished successfully!", logger)
        elif res.status_code == HTTPStatus.NO_CONTENT:
            log_message("Upload successful, content was identical and no new commit was created", logger)
        elif 200 < res.status_code < 300:
            log_message(f"Got unknown successful status code {res.status_code}")
        else:
            raise determine_upload_api_error(res)

    def _poll_mirror_up_to_date(self):
        """
        Synchronization lock for the mirrored repository uploading
        Since the upload is being done "through" DagsHub,
            and we rely on the change first showing up on the original repo, then being synced back to us,
            there is a possibility of uploading a new batch before the sync has been completed,
            during which the DagsHub repo is out of date with the original mirror.
        This is a client-side fix made to mitigate this.
        We poll for the change in the revision and compare it to the revision we had when we last uploaded
        When the revision changes, that means the sync has been completed and we can upload a new batch
        """
        if not self._api.is_mirror:
            return

        # Initial state - assume we can upload
        # Also can upload if last upload didn't have any changes
        if not self._uploading_to_new_branch and (
            self._last_upload_revision is None or not self._last_upload_had_changes
        ):
            return

        poll_interval = 1.0  # seconds
        poll_timeout = 600.0
        start_time = time.time()

        if self.current_progress is not None:
            task = self.current_progress.add_task("Waiting for the mirror to sync", total=None)

            def finish():
                self.current_progress.remove_task(task)

        else:
            status = rich.status.Status("Waiting for the mirror to sync", console=rich_console)
            status.start()

            def finish():
                status.stop()

        while time.time() - start_time < poll_timeout:
            try:
                new_revision = self._api.last_commit_sha(self.branch)
            except BranchNotFoundError:
                # New branch that didn't get back to DagsHub yet
                time.sleep(poll_interval)
                continue
            if new_revision == self._last_upload_revision:
                time.sleep(poll_interval)
            else:
                finish()
                return

        finish()
        logger.warning(
            f"Timed out while polling for a mirror sync finishing after {poll_timeout} s. "
            f"Trying to push anyway, which might not work."
        )

    @property
    def auth(self):
        """
        The auth function is used to authenticate the user with the dagshub API.
            Username and password take priority for authentication, then token.
            If none were provided, it goes through the usual token flow involving the token cache

        :return: The HTTPAuth object

        :meta private:
        """
        if self.username is not None and self.password is not None:
            return httpx.BasicAuth(self.username, self.password)
        if self.token:
            return EnvVarDagshubToken(self.token)
        return dagshub.auth.get_authenticator()

    def directory(self, path: str) -> "DataSet":
        """
        Create a :class:`~dagshub.upload.wrapper.DataSet` object that allows you to stage multiple files before\
        pushing them all to DagsHub in a single commit with :func:`~dagshub.upload.wrapper.DataSet.commit`.

        Args:
            path: The path of the directory in the repository relative to which the files will be uploaded.
        """
        return DataSet(self, path)

    def get_request_url(self, directory, branch=None):
        """
        The get_request_url function returns the URL for uploading a file to DagsHub.

        :param directory: the path to a directory within this repo on DagsHub.
            For example, if you have created your repo in such a
            way that it has two directories named data and models,
            then you could pass one of these strings into this function as an argument.
        :param branch: branch to which upload file.
            Default is None, which uses the branch set on object creation
        :return: The url for uploading a file

        :meta private:
        """
        return self.get_repo_url(CONTENT_UPLOAD_URL, directory, branch)

    def get_files_ui_url(self, directory):
        """
        The get_files_ui_url function returns the URL for seeing the uploaded files on DagsHub.

        :param directory: the path to a directory within this repo on DagsHub.
            For example, if you have created your repo in such a
            way that it has two directories named data and models,
            then you could pass one of these strings into this function as an argument.
        :return: The url that you can navigate to in your browser to see the files

        :meta private:
        """
        return self.get_repo_url(FILES_UI_URL, directory)

    def get_repo_url(self, url_format, directory, branch=None):
        if branch is None:
            branch = self.branch
        return urljoin(
            self.host,
            url_format.format(
                owner=self.owner,
                reponame=self.name,
                branch=branch,
                path=urllib.parse.quote(directory, safe=""),
            ),
        )

    def upload_files_to_bucket(
        self, local_path: Path, remote_path: str, max_workers: int = config.upload_threads, **kwargs
    ):
        """
        Upload a file or directory to an S3 bucket, preserving the directory structure.

        :param local_path: Path to the local directory or file to upload
        :param remote_path: The directory path within the S3 bucket
        :param max_workers: The maximum number of threads to use
        """

        s3 = get_repo_bucket_client(self._api.full_name)
        assert len(remote_path) > 0
        # Get rid of the repo name at the beginning of the path
        if remote_path.split("/")[0] == self.name:
            remote_path = remote_path.split("/", 1)[1]
        if local_path.is_file():
            upload_file_to_s3(s3_client=s3, local_path=local_path, bucket_name=self.name, remote_path=remote_path)
        else:
            upload_tasks = []
            # Walk through the local directory
            for root, dirs, files in os.walk(local_path):
                for filename in files:
                    file_path = os.path.join(root, filename)
                    relative_path = os.path.relpath(file_path, local_path)
                    s3_path = os.path.join(remote_path, relative_path)
                    s3_path = s3_path.replace(os.path.sep, "/")  # Ensure correct path separator for S3
                    upload_tasks.append((file_path, s3_path))

            # Use ThreadPoolExecutor to upload files in parallel
            progress = get_rich_progress(rich.progress.MofNCompleteColumn(), transient=False)
            task_id = progress.add_task("Uploading files...", total=len(upload_tasks))

            with progress:
                with ThreadPoolExecutor(max_workers=max_workers) as executor:
                    futures = [
                        executor.submit(
                            upload_file_to_s3,
                            s3_client=s3,
                            local_path=task[0],
                            bucket_name=self.name,
                            remote_path=task[1],
                            progress=progress,
                            task=task_id,
                        )
                        for task in upload_tasks
                    ]
                    for future in futures:
                        future.result()  # Wait for all futures to complete


class DataSet:
    """
    Not to be confused with DataEngine's datasets.
    This class represents a folder with files that are going to be uploaded to a repo.
    """

    def __init__(self, repo: Repo, directory: str):
        self.files: Dict[str, FileUploadStruct] = {}
        self.repo = repo
        self.directory = self._clean_directory_name(directory)

    def add(self, file: Union[str, BinaryIO], path: Optional[Union[str, Path]] = None):
        """
        Add a file to upload. The file will not be uploaded unless you call :func:`commit`

        Args:
            file: Path to the file on the filesystem OR the contents of the file.
            path: Where to store the file in the repo.
        """

        path, file = self.get_file(file, path)
        if file is not None:
            if path in self.files:
                log_message(f'File already staged for upload on path "{path}". Overwriting', logger)
            self.files[path] = (path, file)

    def add_dir(self, local_path: str, glob_exclude="", commit_message=None, **upload_kwargs):
        """
        Add *and upload* an entire directory to the DagsHub repository.

        By default, this uploads a dvc folder.

        Args:
            local_path: Local path of the directory to upload.
            glob_exclude: Glob pattern to exclude some files from being uploaded.
            commit_message: Message of the commit with the upload.

        The keyword arguments are passed to :func:`Repo.upload_files() <dagshub.upload.Repo.upload_files>`.
        """
        upload_file_number = 100
        max_batch_file_size = 100 * 1024 * 1024
        file_counter = 0

        total_num_files = 0
        for root, dirs, files in os.walk(local_path):
            for filename in files:
                filepath = Path(root) / filename
                if glob_exclude == "" or fnmatch.fnmatch(filepath.as_posix(), glob_exclude) is False:
                    total_num_files += 1

        progress = rich.progress.Progress(
            rich.progress.SpinnerColumn(),
            *rich.progress.Progress.get_default_columns(),
            rich.progress.MofNCompleteColumn(),
            console=rich_console,
            transient=True,
            disable=config.quiet,
        )
        total_task = progress.add_task("Uploading files...", total=total_num_files)
        self.repo.current_progress = progress

        # If user hasn't specified versioning, then assume we're uploading dvc (this makes most sense for folders)
        if "versioning" not in upload_kwargs:
            upload_kwargs["versioning"] = "dvc"

        upload_kwargs["quiet"] = True

        try:
            with progress:
                for root, dirs, files in os.walk(local_path):
                    if len(files) == 0:
                        continue

                    folder_task = progress.add_task(f"Uploading files from {root}", total=len(files))

                    if commit_message is None:
                        commit_message = upload_kwargs.get("commit_message", f"Commit data points in folder {root}")
                    if "commit_message" in upload_kwargs:
                        del upload_kwargs["commit_message"]

                    file_batches: List[List[Path]] = []
                    current_file_batch: List[Path] = []
                    current_batch_file_size = 0

                    for filename in files:
                        filepath = Path(root) / filename
                        if glob_exclude == "" or fnmatch.fnmatch(filepath.as_posix(), glob_exclude) is False:
                            current_file_batch.append(filepath)
                            current_batch_file_size += os.path.getsize(filepath)

                            if (
                                len(current_file_batch) >= upload_file_number
                                or current_batch_file_size >= max_batch_file_size
                            ):
                                file_batches.append(current_file_batch)
                                current_file_batch = []
                                current_batch_file_size = 0

                    if current_file_batch:
                        file_batches.append(current_file_batch)

                    for batch in file_batches:
                        for filepath in batch:
                            path_in_remote = filepath.relative_to(local_path).as_posix()
                            self.add(file=str(filepath), path=path_in_remote)

                        file_counter += len(self.files)
                        self.commit(commit_message, **upload_kwargs)
                        progress.update(folder_task, advance=len(batch), refresh=True)
                        progress.update(total_task, completed=file_counter, refresh=True)

                    progress.remove_task(folder_task)

            log_message(
                f"Directory upload complete, uploaded {file_counter} files"
                f" to {self.repo.get_files_ui_url(self.directory)}",
                logger,
            )
        finally:
            self.repo.current_progress = None

    @staticmethod
    def _clean_directory_name(directory: str):
        """
        The _clean_directory_name function takes a directory name as an argument
        and returns the normalized path of that directory.
        For example, if the input is ../../../ then it will return /. If the input is ./ then it will return ..
        If there are any other characters in the string,
        they will be ignored and only alphanumeric characters (a-zA-Z0-9)
        will be kept.

        :param directory: Specify the directory that will be cleaned
        :return: The normalized path of the directory
                 (The directory name with the path separator normalized to a forward slash)

        """

        return posixpath.normpath(directory)

    @staticmethod
    def get_file(file: Union[str, BinaryIO], path: Optional[Union[str, PurePath]] = None) -> FileUploadStruct:
        """
        The get_file function is a helper function that takes in either a string or an IOBase object and returns
        a tuple containing the file's name and the file itself. If no path is provided, it will default to the name of
        the file.

        :param file: File to upload
        :param path: Desired path of the file in the repository
        :return: A tuple of the path and a file object

        :meta private:
        """

        try:
            str_path: str
            # if path is not provided, fall back to the file name
            if path is None:
                if not isinstance(file, str):
                    raise ValueError(
                        "Could not interpret your file's name. Please specify it in the keyword parameter 'path'."
                    )
                str_path = PurePosixPath(file).name
            else:
                str_path = PurePosixPath(path).as_posix()

            if isinstance(file, str):
                try:
                    f = open(file, "rb")
                    return str_path, f
                except IsADirectoryError:
                    raise IsADirectoryError("'file' must describe a file, not a directory.")

            return str_path, file

        except Exception as e:
            logger.error(e)
            raise

    def _reset_dataset(self):
        """
        The _reset_dataset function clears the files attribute of a Dataset object.

        :return: None
        """
        for f in self.files.values():
            try:
                f[1].close()
            except Exception as e:
                logger.warning(f"Error closing file {f[0]}: {e}")
        self.files.clear()

    def commit(self, commit_message=DEFAULT_COMMIT_MESSAGE, *args, **kwargs):
        """
        Commit files added with :func:`add` to the repo

        Args:
            commit_message: Message of the commit with the upload.

        Other positional and keyword arguments are passed to
        :func:`Repo.upload_files() <dagshub.upload.Repo.upload_files>`
        """

        file_list = list(self.files.values())
        self.repo.upload_files(file_list, self.directory, commit_message=commit_message, *args, **kwargs)
        self._reset_dataset()



import base64
import datetime
import functools
import gzip
import types
import logging
import importlib
from pathlib import PurePath
from typing import Union, TypeVar, Dict, Optional
from urllib.parse import urljoin, quote

logger = logging.getLogger(__name__)


def multi_urljoin(*parts):
    """Shoutout to https://stackoverflow.com/a/55722792"""
    return urljoin(parts[0] + "/", "/".join(quote(part.strip("/"), safe="/") for part in parts[1:]))


def exclude_if_none(value):
    """For skipping serializing None values in dataclasses_json"""
    return value is None


def exclude_if_all_fields_are_none(dataclass_dict: Optional[Dict]):
    if dataclass_dict is None:
        return True
    return all(value is None for value in dataclass_dict.values())


def to_timestamp(ts: Union[float, int, datetime.datetime]) -> int:
    """
    Converts datetime objects or any timestamp-likes into an integer timestamp used by DagsHub
    """
    if isinstance(ts, datetime.datetime):
        return int(ts.timestamp())
    else:
        return int(ts)


def removeprefix(val: str, prefix: str) -> str:
    if val.startswith(prefix):
        return val[len(prefix) :]
    return val


def lazy_load(module_name, source_package=None, callback=None):
    if source_package is None:
        # TODO: need to have a map for commonly used imports here. Also handle dots
        source_package = module_name
    return LazyModule(module_name, source_package, callback)


pathSubclass = TypeVar("pathSubclass", bound=PurePath)


def is_path_relative_to(path_a: pathSubclass, path_b: pathSubclass) -> bool:
    # Polyfill of Path.is_relative
    try:
        path_a.relative_to(path_b)
        return True
    except ValueError:
        return False


class LazyModule(types.ModuleType):
    """Proxy module that lazily imports the underlying module the first time it
    is actually used.

    Shoutout to voxel51 for this :)

    Args:
        module_name: the fully-qualified module name to import
        callback (None): a callback function to call before importing the
            module
    """

    def __init__(self, module_name, source_package, callback=None):
        super().__init__(module_name)
        self._module = None
        self._callback = callback
        self._source_package = source_package

    def __getattr__(self, item):
        if self._module is None:
            self._import_module()

        return getattr(self._module, item)

    def __dir__(self):
        if self._module is None:
            self._import_module()

        return dir(self._module)

    def _import_module(self):
        # Execute callback, if any
        if self._callback is not None:
            self._callback()

        # Actually import the module
        try:
            module = importlib.import_module(self.__name__)
            self._module = module
        except ModuleNotFoundError:
            logger.warning(f"Could not import module {self.__name__}. Make sure to pip install {self._source_package}")
            raise

        # Update this object's dict so that attribute references are efficient
        # (__getattr__ is only called on lookups that fail)
        self.__dict__.update(module.__dict__)


def deprecated(additional_message=""):
    """
    Decorator to mark functions as deprecated. It will print a warning
    message when the function is called.
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            additional = "\n" + additional_message if additional_message else ""
            logger.warning(
                f"DagsHub Deprecation Warning: "
                f"{func.__name__} is deprecated and may be removed in future versions.{additional}",
            )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def wrap_bytes(val: bytes) -> str:
    """
    Handles bytes values for uploading metadata
    The process is gzip -> base64

    :meta private:
    """
    compressed = gzip.compress(val)
    return base64.b64encode(compressed).decode("utf-8")



import datetime
from typing import Optional


def _get_datetime_utc_offset(t: datetime.datetime) -> Optional[str]:
    """
    return a timezone offset in the form of "+03:00" or "-03:00"
    """

    if t.tzinfo is None:
        return None

    offset = t.utcoffset()
    if offset is None:
        return None

    # Format the offset as a string
    offset_hours = int(offset.total_seconds() // 3600)
    offset_minutes = int((offset.total_seconds() % 3600) // 60)
    offset_str = f"{offset_hours:+03d}:{offset_minutes:02d}"
    return offset_str



from dagshub.auth import add_oauth_token, get_token
from dagshub.common.api import RepoAPI, UserAPI
from dagshub.common.api.repo import RepoNotFoundError
from dagshub.upload import create_repo

COLAB_REPO_NAME = "dagshub-drive"


def login() -> str:
    """
    Run custom colab-specific flow, which helps with setting up a repository,
    storage of which will be used as an alternative to Google Drive.

    Returns the name of the repository that can be used with colab-related functionality (``<user>/dagshub-drive``)
    """
    try:
        token = get_token(fail_if_no_token=True)
    except RuntimeError:
        add_oauth_token(referrer="colab")
        token = get_token()

    username = UserAPI.get_user_from_token(token).username

    colab_repo = RepoAPI(f"{username}/{COLAB_REPO_NAME}")
    try:
        colab_repo.get_repo_info()
    except RepoNotFoundError:
        create_repo(COLAB_REPO_NAME)
    print(f"Repository {colab_repo.full_name} is ready for use with Colab. Link to the repository:")
    print(colab_repo.repo_url)
    return colab_repo.full_name



import pytest

from dagshub.common.download import download_url_to_bucket_path


@pytest.mark.parametrize(
    "in_url, expected",
    [
        (
            "https://dagshub.com/api/v1/repos/kirill/bucket-repo/storage/raw/s3/dagshub-storage/images/047.png",
            ("s3", "dagshub-storage", "images/047.png"),
        )
    ],
)
def test_bucket_downloader_path_extraction(in_url, expected):
    actual = download_url_to_bucket_path(in_url)
    assert actual == expected


@pytest.mark.parametrize(
    "in_url", ["https://dagshub.com/api/v1/repos/kirill/bucket-repo/raw/main/images/047.png", "https://google.com"]
)
def test_bucket_download_path_extraction_fail(in_url):
    actual = download_url_to_bucket_path(in_url)
    assert actual is None



import datetime
import logging
from dataclasses import dataclass
from os import PathLike
from pathlib import Path
from typing import Optional, Union, List, Dict, Any, Callable, TYPE_CHECKING, Literal, Sequence

from tenacity import Retrying, stop_after_attempt, wait_exponential, before_sleep_log, retry_if_exception_type

from dagshub.common.download import download_files
from dagshub.common.helpers import http_request
from dagshub.data_engine.annotation import MetadataAnnotations
from dagshub.data_engine.client.models import MetadataSelectFieldSchema, DatapointHistoryResult
from dagshub.data_engine.dtypes import MetadataFieldType

if TYPE_CHECKING:
    from dagshub.data_engine.model.datasource import Datasource

_generated_fields: Dict[str, Callable[["Datapoint"], Any]] = {
    "path": lambda dp: dp.path,
    "datapoint_id": lambda dp: dp.datapoint_id,
    "dagshub_download_url": lambda dp: dp.download_url,
}

logger = logging.getLogger(__name__)


@dataclass
class Datapoint:
    datapoint_id: int
    """
    ID of the datapoint in the database
    """
    path: str
    """
    Path of the datapoint, relative to the root of the datasource
    """
    metadata: Dict[str, Any]
    """
    Dictionary with the metadata
    """
    datasource: "Datasource"
    """
    Datasource this datapoint is from
    """

    def __getitem__(self, item):
        gen_field = _generated_fields.get(item)
        if gen_field is not None:
            return gen_field(self)
        return self.metadata[item]

    def __setitem__(self, key, value):
        if isinstance(value, MetadataAnnotations):
            value = value.to_ls_task()
        self.datasource.implicit_update_context.update_metadata(self.path, {key: value})

    def delete_metadata(self, *fields: str):
        """
        Delete metadata from this datapoint.

        The deleted values can be accessed using versioned query with time set before the deletion.

        Args:
            fields: fields to delete
        """
        self.datasource.delete_metadata_from_datapoints([self], list(fields))

    def delete(self, force: bool = False):
        """
        Delete this datapoint.

        - This datapoint will no longer show up in queries.
        - Does not delete the datapoint's file, only removing the data from the datasource.
        - You can still query this datapoint and associated metadata with \
            versioned queries whose time is before deletion time.
        - You can re-add this datapoint to the datasource by uploading new metadata to it with, for example, \
            :func:`Datasource.metadata_context <dagshub.data_engine.model.datasource.Datasource.metadata_context>`. \
            This will create a new datapoint with new id and new metadata records.
        - Datasource scanning will *not* add this datapoint back.

        Args:
            force: Skip the confirmation prompt
        """
        self.datasource.delete_datapoints([self], force=force)

    def save(self):
        """
        Commit changes to metadata done with one or more dictionary assignment syntax usages.
        `Learn more here <https://dagshub.com/docs/use_cases/data_engine/enrich_datasource\
        /#3-enriching-with-with-dictionary-like-assignment>`_.

        Example::

            specific_data_point['metadata_field_name'] = 42
            specific_data_point.save()

        """

        # if in context block, don't _upload_metadata, it will be done at context end
        if not self.datasource.has_explicit_context:
            self.datasource.upload_metadata_of_implicit_context()

    @property
    def download_url(self):
        """
        str: URL that can be used to download the datapoint's file from DagsHub
        """
        return self.datasource.source.raw_path(self)

    @property
    def path_in_repo(self):
        """
        Path of the datapoint in repo

        :rtype: `PurePosixPath <https://docs.python.org/3/library/pathlib.html#pathlib.PurePosixPath>`_
        """
        return self.datasource.source.file_path(self)

    @staticmethod
    def from_gql_edge(edge: Dict, datasource: "Datasource", fields: List[MetadataSelectFieldSchema]) -> "Datapoint":
        res = Datapoint(
            datapoint_id=int(edge["node"]["id"]),
            path=edge["node"]["path"],
            metadata={},
            datasource=datasource,
        )

        float_fields = {f.name for f in fields if f.valueType == MetadataFieldType.FLOAT}
        date_fields = {f.name for f in fields if f.valueType == MetadataFieldType.DATETIME}

        for meta_dict in edge["node"]["metadata"]:
            key = meta_dict["key"]
            value = meta_dict["value"]
            if key in float_fields:
                value = float(value)
            else:
                if key in date_fields:
                    timezone = meta_dict.get("timeZone")
                    value = _datetime_from_timestamp(value / 1000, timezone or "+00:00")
            res.metadata[key] = value
        return res

    def to_dict(self, metadata_keys: Sequence[str]) -> Dict[str, Any]:
        # Set the autogenerated fields
        res_dict = {k: v(self) for k, v in _generated_fields.items()}
        res_dict.update({key: self.metadata.get(key) for key in metadata_keys})
        return res_dict

    def get_blob(self, column: str, cache_on_disk=True, store_value=False) -> bytes:
        """
        Returns the blob stored in a binary column

        Args:
            column: where to get the blob from
            cache_on_disk: whether to store the downloaded blob on disk.
                If you store the blob on disk, then it won't need to be re-downloaded in the future.
                The contents of datapoint[column] will change to be the path of the blob on the disk.
            store_value: whether to store the blob in memory on the field attached to this datapoint,
                which will make its value accessible later using datapoint[column]
        """
        current_value = self.metadata[column]

        if type(current_value) is bytes:
            # Bytes - it's already there!
            return current_value
        if isinstance(current_value, Path):
            # Path - assume the path exists and is already downloaded,
            #   because it's unlikely that the user has set it themselves
            with current_value.open("rb") as f:
                content = f.read()
            if store_value:
                self.metadata[column] = content
            return content

        elif type(current_value) is str:
            # String - This is probably the hash of the blob, get that from dagshub
            blob_url = self.blob_url(current_value)
            blob_location = self.blob_cache_location / current_value

            # Make sure that the cache location exists
            if cache_on_disk:
                self.blob_cache_location.mkdir(parents=True, exist_ok=True)

            content = _get_blob(blob_url, blob_location, self.datasource.source.repoApi.auth, cache_on_disk, True)
            if type(content) is str:
                raise RuntimeError(f"Error while downloading blob: {content}")

            if store_value:
                self.metadata[column] = content
            elif cache_on_disk:
                self.metadata[column] = blob_location

            return content
        else:
            raise ValueError(f"Can't extract blob metadata from value {current_value} of type {type(current_value)}")

    def download_file(
        self, target: Optional[Union[PathLike, str]] = None, keep_source_prefix=True, redownload=False
    ) -> PathLike:
        """
        Downloads the datapoint to the target_dir directory

        Args:
            target: Where to download the file (either a directory, or the full path).\
                If not specified, then downloads to\
                :func:`datasource's default location \
                <dagshub.data_engine.model.datasource.Datasource.default_dataset_location>`.
            keep_source_prefix: If True, includes the prefix of the datasource in the download path.
            redownload: Whether to redownload a file if it exists on the filesystem already.

        .. note::
            We don't do any hashsum checks, so if it's possible that the file has been updated,
            set ``redownload`` to True

        Returns:
            Path to the downloaded file
        """

        target_path = self.datasource.default_dataset_location if target is None else Path(target).expanduser()

        # Check if the specified path looks like a file
        # by checking if there's an extension, or it's an already existing file
        n = target_path.name
        target_is_already_file = (target_path.exists() and target_path.is_file()) or (
            "." in n and not n.startswith(".")
        )

        if not target_is_already_file:
            if keep_source_prefix:
                target_path = target_path / self.path_in_repo
            else:
                target_path = target_path / self.path

        download_files([(self.download_url, target_path)], skip_if_exists=not redownload)
        return target_path

    @property
    def blob_cache_location(self):
        return self.datasource.default_dataset_location / ".metadata_blobs"

    def blob_url(self, sha):
        return self.datasource.source.blob_path(sha)

    def get_version_timestamps(
        self,
        fields: Optional[List[str]] = None,
        from_time: Optional[datetime.datetime] = None,
        to_time: Optional[datetime.datetime] = None,
    ) -> List[DatapointHistoryResult]:
        """
        Get the timestamps of all versions of this datapoint, where the specified fields have changed.

        Args:
            fields: List of fields to check for changes. If None, all fields are checked.
            from_time: Only search versions since this time. If None, the start time is unbounded
            to_time: Only search versions until this time. If None, the end time is unbounded

        Returns:
            List of objects with information about the versions.
        """
        history = self.datasource.source.client.get_datapoint_history([self], fields, from_time, to_time)
        return history.get(self.path, [])


def _get_blob(
    url: Optional[str],
    cache_path: Optional[Path],
    auth,
    cache_on_disk: bool,
    return_blob: bool,
    path_format: Literal["str", "path"] = "path",
) -> Optional[Union[Path, str, bytes]]:
    """
    Args:
        url: url to download the blob from
        cache_path: where the cache for the blob is (laods from it if exists, stores there if it doesn't)
        auth: auth to use for getting the blob
        cache_on_disk: whether to store the downloaded blob on disk. If False we also turn off the cache checking
        return_blob: if True returns the blob of the downloaded data, if False returns the path to the file with it
    """
    if url is None:
        return None
    assert cache_path is not None

    if cache_on_disk and cache_path.exists():
        if return_blob:
            with cache_path.open("rb") as f:
                return f.read()
        else:
            if path_format == "str":
                cache_path = str(cache_path)
            return cache_path

    def get():
        resp = http_request("GET", url, auth=auth)
        if 200 <= resp.status_code < 300:
            return resp.content
        elif resp.status_code == 404:
            raise Exception(f"Blob not found at {url}")
        elif resp.status_code > 400:
            raise RuntimeError(f"Got status code {resp.status_code} from server")
        else:
            raise Exception(f"Non-retrying status code {resp.status_code} returned")

    try:
        for attempt in Retrying(
            retry=retry_if_exception_type(RuntimeError),
            stop=stop_after_attempt(5),
            wait=wait_exponential(multiplier=1, min=4, max=10),
            before_sleep=before_sleep_log(logger, logging.WARNING),
        ):
            with attempt:
                content = get()
    except Exception as e:
        return f"Error while downloading binary blob: {e}"

    if cache_on_disk:
        with cache_path.open("wb") as f:
            f.write(content)

    if return_blob:
        return content
    else:
        if path_format == "str":
            cache_path = str(cache_path)
        return cache_path


def _datetime_from_timestamp(timestamp, utc_offset):
    offset_hours, offset_minutes = map(int, utc_offset.split(":"))
    offset = datetime.timedelta(hours=offset_hours, minutes=offset_minutes)

    tz = datetime.timezone(offset)

    return datetime.datetime.fromtimestamp(timestamp).astimezone(tz)



import platform
import shutil
import subprocess
import configparser
import logging
import os
from pathlib import Path
from typing import Tuple, Optional, Union

from ..common.helpers import log_message, prompt_user
from ..auth import get_token
import dagshub.common.config

logger = logging.getLogger(__name__)


def check_and_provide_install_script(quiet=False):
    """
    Checks whether Rclone and fuse3 are installed on the system. If not, it provides the necessary
    installation commands.

    :param quiet: Optional. A boolean flag that controls the output of the function. If False, the function will
    print messages about its operation.
    No parameters are required for this function.
    """
    packages = {
        "rclone": "sudo -v && curl https://rclone.org/install.sh | sudo bash",
        "fusermount3": "apt install -y fuse3",
    }
    missing_packages = []

    for package, install_command in packages.items():
        try:
            # Check if the package is installed by querying its version
            subprocess.run(
                [package, "--version"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                check=True,
            )
        except (subprocess.CalledProcessError, FileNotFoundError):
            # If not installed, add the package to the list of missing packages
            missing_packages.append((package, install_command))

    if missing_packages:
        # Format a string listing all missing packages
        missing_packages_list = ", ".join([pkg for pkg, _ in missing_packages])
        if not quiet:
            log_message(f"The following packages are not installed: {missing_packages_list}.")
        if platform.system().lower() == "linux" and shutil.which("apt") is not None:
            response = prompt_user("Do you want to install all missing packages?")

            if response:
                for package, cmd in missing_packages:
                    log_message(f"Installing {package}...")
                    try:
                        subprocess.run(cmd, shell=True, check=True)
                        log_message(f"{package} installed successfully.")
                    except subprocess.CalledProcessError as e:
                        logger.error(f"Failed to install {package}: {e}")
                        # Non-Linux OS or 'apt' not available
                        log_message("Manual installation required. Please install Rclone and fuse3 for your system.")
                        log_message("Rclone installation instructions: https://rclone.org/install/")
                        log_message(
                            "For fuse3, please consult your operating system's package manager or " "documentation."
                        )
            else:
                log_message("Skipping installation of missing packages.")
        else:
            # Non-Linux OS or 'apt' not available
            log_message("Manual installation required. Please install Rclone and fuse3 for your system.")
            log_message("Rclone installation instructions: https://rclone.org/install/")
            log_message("For fuse3, please consult your operating system's package manager or documentation.")
            return
    else:
        if not quiet:
            log_message("All packages are installed.")


def rclone_init(repo_owner: str, conf_path: Optional[Path] = None, update=False, quiet=False) -> Tuple[str, Path]:
    """
    Initializes or updates the Rclone configuration for a DagsHub repository.

    :param repo_owner: The owner of the repository. This is used to create a unique section in the Rclone configuration.
    :param conf_path: Optional. The path to the Rclone configuration file. If not provided, the default path is used.
    :param update: Optional. A boolean flag indicating whether to update the configuration if it already exists.
        Defaults to False.
    :param quiet: Optional. A boolean flag that controls the output of the function. If False, the function will
        print messages about its operation.
    :return: Name of the remote for rclone + The absolute path to the Rclone configuration file.
    """
    # Make sure RClone and fuse3 are properly installed
    check_and_provide_install_script(quiet=True)

    if conf_path is None:
        root_path = Path.home() / ".config/rclone/"
        conf_path = root_path / "rclone.conf"
        if not os.path.exists(root_path):
            os.makedirs(Path.home() / ".config/rclone/")
    else:
        conf_path = Path(conf_path)

    token = get_token()

    config = configparser.ConfigParser()
    if conf_path.exists():
        config.read(conf_path)

    # Determine the section name
    if "dagshub" in config and not update:
        remote_name = f"dagshub-{repo_owner}"
    else:
        remote_name = "dagshub"

    config[remote_name] = {
        "type": "s3",
        "provider": "Other",
        "access_key_id": token,
        "secret_access_key": token,
        "endpoint": f"{dagshub.common.config.host}/api/v1/repo-buckets/s3/{repo_owner}",
    }

    with conf_path.open("w") as f:
        config.write(f)

    if not quiet:
        # Inform the user about the remote name
        log_message(f"Configuration complete. The remote '{remote_name}' has been created/updated in '{conf_path}'.")
        log_message(
            f"Example usage with rclone: `rclone ls {remote_name}:<your-bucket-name>` to list the contents of "
            f"'your-bucket-name'."
        )

    return remote_name, conf_path.absolute()


def sync(repo: str, local_path: Union[str, os.PathLike], remote_path: Union[str, os.PathLike]):
    """
    Synchronizes the contents of a local directory with a specified remote directory in a DagsHub repository using
    Rclone.

    :param repo: A string in the ``<repo_owner>/<repo_name>`` format representing the target DagsHub repository.
    :param local_path: A Path object or string pointing to the local directory to be synchronized.
    :param remote_path: A Path object or string representing the remote directory path relative to the DagsHub Storage
        bucket root.
    """
    # Extract repo_owner and repo_name from repo argument
    repo_owner, repo_name = repo.split("/")

    # Ensure the repository is configured in Rclone
    remote_name, conf_path = rclone_init(repo_owner=repo_owner, quiet=True)

    # Convert local_path to a string, in case it's a Path object
    local_path_str = str(local_path)

    try:
        # Construct the Rclone sync command
        sync_command = [
            "rclone",
            "sync",
            "--ignore-existing",
            local_path_str,
            f"{remote_name}:{repo_name}/{remote_path}",
            "--progress",
        ]

        # Execute the Rclone sync command with Popen and stream output
        with subprocess.Popen(
            sync_command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            bufsize=1,
            universal_newlines=True,
        ) as p:
            for line in p.stdout:
                log_message(line)

        log_message(f"Successfully synchronized {local_path_str} to DagsHub Storage {remote_path}.")
    except subprocess.CalledProcessError as e:
        log_message(f"Failed to synchronize {local_path_str} to DagsHub Storage {remote_path}: {e}")


def mount(repo: str, cache: bool = False, path: Path = None) -> os.PathLike:
    """
    Mounts a DagsHub repository bucket to a local directory.

    .. warning::
        This function is only supported on Linux machines and on macOS via FUSE for macOS (FUSE-T or macFUSE).
        It may not work as expected on other operating systems due to differences in the handling of filesystem mounts.

    :param repo: The repository in the format ``<repo_owner>/<repo_name>``. This is used to determine
        the remote name and mount point.

    :param cache: Optional. A boolean flag that enables or disables the cache feature. If True, caching is enabled with
        specific settings `--vfs-cache-max-age 24h`.

    :param path: Optional. A Path object specifying the custom mount path. If not provided, the mount directory is
        determined based on the current working directory and the repository name.

    .. note::
        This function, as well :func:`sync` are using `RClone <https://rclone.org/>`_ to work.
        If it's not installed, you'll get instructions how to install it.
    """
    # Parse the repo string to get the repo owner and name
    repo_owner, repo_name = repo.split("/")

    # Determine the mount directory
    if path is not None:
        mount_dir = Path(path)
    else:
        current_dir = Path(os.getcwd()).name
        if current_dir == repo_name:
            mount_dir = Path("dagshub_storage")
        else:
            mount_dir = Path(repo_name) / "dagshub_storage"

    # 2. Configure the rclone conf
    remote_name, conf_path = rclone_init(repo_owner=repo_owner, quiet=True)

    # 3. Prepare and execute the mount command
    try:
        mount_dir.mkdir(parents=True, exist_ok=True)
    except OSError:
        # In Colab, when a cell fails to run, all mounts become broken, and os level actions like mkdir fail,
        # with an OSError: Transport Endpoint Disconnected. To fix this you must unmount the bucket and remount it,
        # so here we are attempting to handle this on behalf of the user. If it's a different error,
        # it will fail the second time.
        logger.error("Encountered OSError, attempting to remount...")
        unmount(repo)
        mount_dir.mkdir(parents=True, exist_ok=True)

    mount_command = [
        "rclone",
        "mount",
        "--daemon",
        "--allow-other",
        "--vfs-cache-mode",
        "full",
        f"{remote_name}:{repo_name}/",
        f"{mount_dir}/",
    ]

    # Add cache options if needed
    if cache:
        mount_command.extend(["--vfs-cache-max-age", "24h"])

    try:
        # Execute the mount command
        subprocess.run(mount_command, check=True)
        log_message(f"Successfully mounted DagsHub Storage in '{repo_name}' to '{mount_dir}'.")
        log_message(f'To unmount, run `dagshub.storage.unmount(repo="{repo}", path="{mount_dir}")`.')
        return mount_dir
    except subprocess.CalledProcessError as e:
        logger.error(f"Failed to mount DagsHub Storage in '{repo_name}': {e}")


def unmount(repo, path=None):
    """
    Unmounts a previously mounted DagsHub repository bucket from the local file system.

    :param repo: The name of the repository. Used to determine the default mount point if a custom path is
        not provided.
    :param path: Optional. A custom path to the mount point. If not provided, the default logic is used to determine
        the mount point based on the repository name.
    """
    repo_parts = repo.split("/")
    repo_name = repo_parts[-1]

    if path:
        # If a custom path is provided, use it as the mount point
        mount_point = Path(path)
    else:
        # Default logic to determine the mount point
        current_dir = Path(os.getcwd()).name
        if current_dir == repo_name:
            mount_point = Path("dagshub_storage")
        else:
            mount_point = Path(repo_name) / "dagshub_storage"

    try:
        subprocess.run(["fusermount", "-u", str(mount_point)], check=True)
        log_message(f"Successfully unmounted '{mount_point}'.")
    except subprocess.CalledProcessError as e:
        logger.error(f"Failed to unmount '{mount_point}': {e}")



import rich.console
from .notebook import is_inside_colab, is_inside_notebook


force_terminal = False if is_inside_notebook() else None

rich_console = rich.console.Console(force_terminal=force_terminal)

__all__ = [
    is_inside_colab.__name__,
    is_inside_notebook.__name__,
    "rich_console",
    "force_terminal",
]



__version__ = "0.5.10"
from .logger import DAGsHubLogger, dagshub_logger
from .common.init import init
from .upload.wrapper import upload_files
from . import notebook
from .repo_bucket import get_repo_bucket_client
from .ls_client import get_label_studio_client
from . import storage

__all__ = [
    DAGsHubLogger.__name__,
    dagshub_logger.__name__,
    init.__name__,
    upload_files.__name__,
    notebook.save_notebook.__name__,
    get_repo_bucket_client.__name__,
    get_label_studio_client.__name__,
    storage.__name__,
]



import hashlib
from contextlib import nullcontext
from typing import Optional, NamedTuple
import logging
import urllib
import uuid
import httpx
import webbrowser

from dagshub.auth.token_auth import OAuthDagshubToken
from dagshub.common import config, rich_console

logger = logging.getLogger(__name__)


def oauth_flow(host: str, client_id: Optional[str] = None, referrer: Optional[str] = None) -> OAuthDagshubToken:
    """
    Initiate the OAuth 2.0 flow for obtaining an access token.

    Args:
        host (str): The URL of the OAuth provider.
        client_id (Optional[str], optional): The client ID used for authentication.
            If not provided, it will use the default client ID from the configuration.
        referrer (Optional[str], optional): For custom referral

    Returns:
        OAuthDagshubToken: The authenticated token object.
    """

    flow_data = oauth_flow_url(host, client_id, referrer)

    webbrowser.open(flow_data.auth_link)

    rich_console.print(
        "[bold]:exclamation::exclamation::exclamation: AUTHORIZATION REQUIRED "
        ":exclamation::exclamation::exclamation:[/bold]",
        justify="center",
    )
    # Doing raw prints here because the rich syntax breaks in colab
    # Printing them line by line, because the link has to be its own print in order for Colab parser to correctly parse
    # the whole link
    print("\n\nOpen the following link in your browser to authorize the client:")
    print(flow_data.auth_link)
    print("\n")

    token = oauth_flow_post(flow_data)
    logger.debug(f"Got token: {token}")
    return token


class OAuthFlowData(NamedTuple):
    """Data object containing all necessary information for OAuth flow."""
    auth_link: str
    """
    The link the user needs to visit to confirm the OAuth flow.
    This link should be shown to the user, who needs to open it in their web browser.
    """
    # Other fields needed to complete the flow in oauth_flow_post
    client_id: str
    oauth_url: str
    middle_man_request_id: str
    state: uuid.UUID


def oauth_flow_url(host: str, client_id: str, referrer: str) -> OAuthFlowData:
    """
    Generate OAuth flow URL and required data for the authentication flow.
    You can use this together with :func:`oauth_flow_post` to complete the full flow manually in custom UIs.

    Args:
        host (str): The URL of the OAuth provider.
        client_id (str): The client ID used for authentication.
        referrer (str): Custom referrer URL, if any.

    Returns:
        OAuthFlowData: Object containing all data needed for the OAuth flow.
    """
    host = host.strip("/")
    oauth_url = urllib.parse.urljoin(host, "login/oauth")
    client_id = client_id or config.client_id
    state = uuid.uuid4()
    middle_man_request_id = hashlib.sha256(uuid.uuid4().bytes).hexdigest()
    auth_link = (
        f"{oauth_url}/authorize?state={state}&client_id={client_id}&middleman_request_id={middle_man_request_id}"
    )
    if referrer is not None:
        auth_link += f"&referrer={referrer}"

    return OAuthFlowData(
        auth_link=auth_link,
        client_id=client_id,
        oauth_url=oauth_url,
        middle_man_request_id=middle_man_request_id,
        state=state
    )


def oauth_flow_post(flow_data: OAuthFlowData, quiet: bool = False) -> OAuthDagshubToken:
    """
    Complete the OAuth flow by retrieving an access token (token exchange).
    Blocks until the user completes the flow via the browser, or times out.

    Args:
        flow_data (OAuthFlowData): The flow data containing all required authentication parameters.

    Returns:
        OAuthDagshubToken: The authenticated token.
        :param quiet: Use True to suppress print output
    """
    with (rich_console.status("Waiting for authorization") if not quiet else nullcontext()):
        res = httpx.post(
            f"{flow_data.oauth_url}/middleman",
            data={"request_id": flow_data.middle_man_request_id},
            timeout=None
        )
    if res.status_code != 200:
        raise Exception(f"Error while getting OAuth code: HTTP {res.status_code}, Body: {res.json()}")
    code = res.json()
    res = httpx.post(
        f"{flow_data.oauth_url}/access_token",
        data={
            "client_id": flow_data.client_id,
            "code": code,
            "state": str(flow_data.state)
        }
    )
    if res.status_code != 200:
        raise Exception(f"Error while getting OAuth token: HTTP {res.status_code}, Body: {res.json()}")
    token = OAuthDagshubToken.deserialize(res.json())
    return token


# TODO: What should this be used for? Not currently raised anywhere
class OauthNonInteractiveShellException(Exception):
    pass



import configparser
import os
import urllib
import urllib.parse
from os.path import exists
from pathlib import Path
from typing import Optional

from dagshub.auth import get_token
from dagshub.common import config
from dagshub.common.api import RepoAPI, UserAPI
from dagshub.common.api.repo import RepoNotFoundError
from dagshub.common.determine_repo import determine_repo
from dagshub.common.helpers import log_message
from dagshub.upload import create_repo
from dagshub.mlflow import patch_mlflow as _patch_mlflow

from dagshub.common.util import lazy_load

git = lazy_load("git")


def init(
    repo_name: Optional[str] = None,
    repo_owner: Optional[str] = None,
    url: Optional[str] = None,
    root: Optional[str] = None,
    host: Optional[str] = None,
    mlflow: bool = True,
    dvc: bool = False,
    patch_mlflow: bool = False,
):
    """
    Initialize a DagsHub repository or DagsHub-related functionality.

    Initialization includes:
        1) Creates a repository on DagsHub if it doesn't exist yet.
        2) If ``dvc`` flag is set, adds the DagsHub repository as a dvc remote.
        3) If ``mlflow`` flag is set, initializes MLflow environment variables to enable logging experiments into the\
            DagsHub hosted MLflow. That means that if you call ``dagshub.init()`` in your script,\
            then any MLflow function called later in the script will log to the DagsHub hosted MLflow.

    Arguments:
        repo_name: Along with ``repo_owner`` defines the repository on DagsHub.
        repo_owner: Along with ``repo_name`` defines the repository on DagsHub.
        url: Url to the repository on DagsHub. Can be used as an alternative to ``repo_owner/repo_name`` arguments.
        root: Path to the locally hosted git repository.
            If it's not set, tries to find a repository traversing up the filesystem.
        host: Address of the DagsHub instance with the repository.
            If specified, then all request to dagshub are going to be sent to this instance.
        mlflow: Configure MLflow to log experiments to DagsHub.
        dvc: Configure a dvc remote in the repository.
        patch_mlflow: Run :func:`dagshub.mlflow.patch_mlflow` so errors while logging with MLflow don't stop execution
    """
    if host is None:
        host = config.host
    else:
        config.set_host(host)

    if root is None:
        root = os.path.abspath(".")

    # URL specified - ignore repo name and owner args, prioritize url over it
    if url is not None:
        repo_owner, repo_name = None, None

    if None in [repo_owner, repo_name] and (repo_owner is not None or repo_name is not None):
        raise AttributeError("Both repo_owner and repo_name should be set")

    # Build URL from repo owner and name
    if repo_owner and repo_name:
        url = urllib.parse.urljoin(f"{host}/", f"{repo_owner}/{repo_name}")
    else:
        if not url:
            # Try to get the url of the repo from the git repo
            repo, branch = determine_repo(root)
            url = repo.repo_url

        if url.endswith(".git"):
            url = url[:-4]
        # Extract the owner and name from the repo_url
        parts = url.split("/")
        repo_owner, repo_name = parts[-2], parts[-1]

    # Create the repo if it wasn't created
    repo_api = RepoAPI(f"{repo_owner}/{repo_name}", host=host)
    try:
        repo_api.get_repo_info()
    except RepoNotFoundError:
        log_message(f"Repository {repo_name} doesn't exist, creating it under current user.")
        create_repo(repo_name)

    # Get the token for the configs
    token = get_token(host=host)

    # Configure MLFlow
    if mlflow:
        os.environ["MLFLOW_TRACKING_URI"] = f"{url}.mlflow"
        os.environ["MLFLOW_TRACKING_USERNAME"] = UserAPI.get_user_from_token(token).username
        os.environ["MLFLOW_TRACKING_PASSWORD"] = token

        log_message(f'Initialized MLflow to track repo "{repo_owner}/{repo_name}"')

    if patch_mlflow:
        _patch_mlflow()

    # Configure DVC
    if dvc:
        git_repo = git.Repo(root, search_parent_directories=True)
        git_repo_path = Path(git_repo.git_dir)

        Path(git_repo_path / ".dvc").mkdir(parents=True, exist_ok=True)
        write = True

        dvc_config = configparser.ConfigParser()
        dvc_config_local = configparser.ConfigParser()
        dvc_config.read(git_repo_path / ".dvc" / "config")
        dvc_config_local.read(git_repo_path / ".dvc" / "config.local")

        for section in dvc_config.sections():
            if "url" in dvc_config[section] and host in dvc_config[section]["url"]:
                write = False

        remote = "dagshub" if "origin" in dvc_config.sections() else "origin"
        if write:
            dvc_config_local[f"'remote \"{remote}\"'"] = {"auth": "basic", "user": token, "password": token}
            dvc_config[f"'remote \"{remote}\"'"] = {"url": f"{url}.dvc"}

            with open(git_repo_path / ".dvc" / "config", "w") as config_file, open(
                git_repo_path / ".dvc" / "config.local", "w"
            ) as config_local_file:
                dvc_config.write(config_file)
                dvc_config_local.write(config_local_file)
                log_message(f'Added new remote "{remote}" with url = {url}')

        if not exists(git_repo_path / ".dvc" / ".gitignore"):
            with open(git_repo_path / ".dvc" / ".gitignore", "w") as config_gitignore:
                config_gitignore.write(config.CONFIG_GITIGNORE)

    log_message(f"Repository {repo_owner}/{repo_name} initialized!")



import inspect
import logging
import sys
import threading
import traceback
from types import ModuleType, FunctionType
from typing import Callable, Optional, Dict, Set, List, Tuple

from tenacity import Retrying, stop_after_delay

logger = logging.getLogger(__name__)

_top_level_functions = [
    "mlflow.log_artifact",
    "mlflow.log_artifacts",
    "mlflow.log_dict",
    "mlflow.log_figure",
    "mlflow.log_image",
    "mlflow.log_input",
    "mlflow.log_metric",
    "mlflow.log_metrics",
    "mlflow.log_param",
    "mlflow.log_params",
    "mlflow.log_table",
    "mlflow.log_text",
]

_mlflow_client_functions = [
    "mlflow.tracking.client.MlflowClient.log_artifact",
    "mlflow.tracking.client.MlflowClient.log_artifacts",
    "mlflow.tracking.client.MlflowClient.log_batch",
    "mlflow.tracking.client.MlflowClient.log_dict",
    "mlflow.tracking.client.MlflowClient.log_figure",
    "mlflow.tracking.client.MlflowClient.log_image",
    "mlflow.tracking.client.MlflowClient.log_inputs",
    "mlflow.tracking.client.MlflowClient.log_metric",
    "mlflow.tracking.client.MlflowClient.log_param",
    "mlflow.tracking.client.MlflowClient.log_table",
    "mlflow.tracking.client.MlflowClient.log_text",
]

_default_patch_functions = _top_level_functions + _mlflow_client_functions

_default_guaranteed_raises = ["log_model"]


class MlflowMonkeyPatch:
    context_manager_functions = ["start_run"]

    def __init__(self, funcs_to_patch: List[str], guaranteed_raises: List[str]):
        self.funcs_to_patch = set(funcs_to_patch)
        # Functions that guarantee to raise an exception if they are in the stack
        # Example: all mlflow.<framework>.log_model functions underneath use log_artifact,
        # but we might want to raise exceptions, so logging models doesn't get swallowed
        self.functions_with_guaranteed_raises = set(guaranteed_raises)

        self.original_func_lookup: Dict[str, Callable] = {}
        self.patched_modules: Set[str] = set()

    def _wrap_func(self, fn: Callable) -> Callable:
        from mlflow.exceptions import MlflowException

        def is_top_level_mlflow_call() -> bool:
            tb = traceback.extract_stack()

            # Count the times `dh_mlflow_wrap_fn` functions show up in the stack trace,
            # if it's only once - means it's the top-most level wrapped mlflow call

            wrap_fn_met = False
            for stack_line in tb:
                # If there is a guaranteed raise in the stack - return False
                if stack_line.name in self.functions_with_guaranteed_raises:
                    return False
                is_wrap_func = stack_line.filename == __file__ and stack_line.name in {
                    "dh_mlflow_wrap_fn",
                    "dh_mlflow_context_manager_wrap_fn",
                }
                if not is_wrap_func:
                    continue
                elif wrap_fn_met:
                    return False
                else:
                    wrap_fn_met = True

            return True

        def dh_mlflow_wrap_fn(*args, **kwargs):
            try:
                return fn(*args, **kwargs)
            except MlflowException:
                # Allow exceptions to bubble up if it's not the top-level user made call
                if not is_top_level_mlflow_call():
                    raise
                logger.exception(
                    f"MlflowException occurred when calling {fn.__name__} "
                    f"but was suppressed due to running dagshub.mlflow.patch_mlflow:"
                )

        def dh_mlflow_context_manager_wrap_fn(*args, **kwargs):
            # Context managers (start_run), always return a value, so those we retry for a long amount of time instead
            # Swallowing this function's exception would break any call after that
            # Alternative might be returning a mock object that no-ops any calls to it
            for attempt in Retrying(stop=stop_after_delay(60 * 15), reraise=True):
                with attempt:
                    return fn(*args, **kwargs)

        if fn.__name__ in self.context_manager_functions:
            return dh_mlflow_context_manager_wrap_fn
        else:
            return dh_mlflow_wrap_fn

    def is_function_already_patched(self, func_name: str) -> bool:
        return func_name in self.original_func_lookup

    def monkey_patch_module(self, module: ModuleType):
        from mlflow.utils.lazy_load import LazyLoader

        # Add module into set of patched modules, so there's no circular import recursion
        self.patched_modules.add(module.__name__)

        # Traverses the module tree and monkey patches the accessible functions with wrap_func
        for attr_name in dir(module):
            # Skip dunder and private methods, bad idea to wrap them, we are only concerned with public-facing API
            if attr_name.startswith("_"):
                continue

            full_attr_name = f"{module.__name__}.{attr_name}"
            attr = getattr(module, attr_name)

            is_function = isinstance(attr, FunctionType)
            is_class = inspect.isclass(attr)

            if is_function:
                if full_attr_name not in self.funcs_to_patch:
                    continue
                if self.is_function_already_patched(full_attr_name):
                    continue
                self.original_func_lookup[full_attr_name] = attr
                setattr(module, attr_name, self._wrap_func(attr))
                continue

            if is_class:
                for class_attr_name in dir(attr):
                    if class_attr_name.startswith("_"):
                        continue
                    class_attr = getattr(attr, class_attr_name)
                    full_class_attr_name = f"{full_attr_name}.{class_attr_name}"
                    if (
                        isinstance(class_attr, FunctionType)
                        and not self.is_function_already_patched(full_class_attr_name)
                        and full_class_attr_name in self.funcs_to_patch
                    ):
                        self.original_func_lookup[full_class_attr_name] = class_attr
                        setattr(attr, class_attr_name, self._wrap_func(class_attr))

            is_module = isinstance(attr, ModuleType)
            # Ignore lazy loaded modules, don't want to load them all at once
            # TODO: maybe let users tune this somehow, although from what I've seen most of those are wrappers,
            #  and the actual mlflow logging functions are eagerly loaded
            is_lazy_module = isinstance(attr, LazyLoader)

            if not is_module or is_lazy_module:
                continue

            full_module_name = attr.__name__

            if not full_module_name.startswith("mlflow"):
                continue

            if full_module_name in self.patched_modules:
                continue

            # Recurse into the module and patch it
            self.monkey_patch_module(attr)

    def unpatch(self):
        logger.warning("Unpatching MLflow to raise exceptions again")

        for unpatched_key, unpatched_func in self.original_func_lookup.items():
            attrs = unpatched_key.split(".")
            module_name = attrs[0]
            attrs = attrs[1:]
            attr = sys.modules[module_name]
            for index, attr_name in enumerate(attrs):
                parent_attr = attr
                attr = getattr(parent_attr, attr_name)
                if index == len(attrs) - 1:
                    setattr(parent_attr, attr_name, unpatched_func)

        self.patched_modules.clear()
        self.original_func_lookup.clear()


_global_mlflow_patch: Optional[MlflowMonkeyPatch] = None
_patch_mutex = threading.Lock()


def _patch_mlflow(funcs_to_patch: List[str], guaranteed_raises: List[str]):
    import mlflow

    global _global_mlflow_patch

    with _patch_mutex:
        if _global_mlflow_patch is not None:
            logger.warning("MLflow is already patched")
            return

        logger.warning(
            "Patching MLflow to prevent any MLflow exceptions from being raised. "
            "Call dagshub.mlflow.unpatch_mlflow() to undo"
        )

        patch = MlflowMonkeyPatch(funcs_to_patch, guaranteed_raises)
        patch.monkey_patch_module(mlflow)
        _global_mlflow_patch = patch


def _resolve_patches(
    include: Optional[List[str]],
    exclude: Optional[List[str]],
    patch_top_level: bool,
    patch_mlflow_client: bool,
    raise_on_log_model: bool,
) -> Tuple[List[str], List[str]]:
    """
    Resolve the arguments for patch_mlflow

    Returns:
        Tuple of:
        - List of functions that need to be patched
        - List of functions that should raise exceptions if they are in the stack
    """
    funcs_to_patch = _default_patch_functions.copy()
    guaranteed_raises = _default_guaranteed_raises.copy()

    if not patch_top_level:
        for f in _top_level_functions:
            if f in funcs_to_patch:
                funcs_to_patch.remove(f)

    if not patch_mlflow_client:
        for f in _mlflow_client_functions:
            if f in funcs_to_patch:
                funcs_to_patch.remove(f)

    if include is not None:
        funcs_to_patch = funcs_to_patch + include

    if exclude is not None:
        for f in exclude:
            if f in funcs_to_patch:
                funcs_to_patch.remove(f)

    if not raise_on_log_model:
        if "log_model" in guaranteed_raises:
            guaranteed_raises.remove("log_model")

    return funcs_to_patch, guaranteed_raises


def patch_mlflow(
    include: Optional[List[str]] = None,
    exclude: Optional[List[str]] = None,
    raise_on_log_model=True,
    patch_top_level=True,
    patch_mlflow_client=True,
):
    """
    Patch MLflow functions, making some of them stop raising exceptions, instead logging them to console.
    This is useful for long runs that have MLflow failing occasionally,
    so you don't have to restart the run if logging failed.

    By default, all top-level ``mlflow.log_...`` and ``MlflowClient.log_...`` functions are patched.

    Args:
        include: List of full names of functions you want to patch additionally to the default ones.
        exclude: List of functions you DON'T want to patch.\
            Use this if you need to make sure, for example, all ``log_artifact``
            functions raise an exception on failure:

            .. code-block:: python

                patch_mlflow(exclude=["mlflow.log_artifact", "mlflow.log_artifacts"])

        raise_on_log_model: If ``True``, patched log calls still raise an exception
            if they are called from a ``log_model`` function.

            This ensures that even if you want to ignore failed calls to ``log_artifact`` or ``log_figure`` etc.,
            failed calls to ``log_model`` in particular will still raise errors.
            This should work for all supported ML frameworks, as long as the function is named ``log_model``.
        patch_top_level: Whether to patch the top level ``mlflow`` functions.
        patch_mlflow_client: Whether to patch the ``MlflowClient`` class.
    """
    funcs_to_patch, guaranteed_raises = _resolve_patches(
        include=include,
        exclude=exclude,
        patch_top_level=patch_top_level,
        patch_mlflow_client=patch_mlflow_client,
        raise_on_log_model=raise_on_log_model,
    )

    _patch_mlflow(funcs_to_patch, guaranteed_raises)


def unpatch_mlflow():
    """
    Removes the failsafe MLflow patching, returning all MLflow functions to their original state
    """
    global _global_mlflow_patch

    with _patch_mutex:
        if _global_mlflow_patch is None:
            logger.warning("MLflow should be unpatched already")
            return
        _global_mlflow_patch.unpatch()
        _global_mlflow_patch = None


# TODO: add a "with_unpatch" (name subject to change) context manager, that disables the patching
# Implementation: set a THREAD LOCAL flag that disables the wrapping behaviour inside of the functions



from multiprocessing import Process
from dagshub.common.util import lazy_load
from typing import TYPE_CHECKING, Union, Any
from dagshub.data_engine.client.loaders.base import DagsHubDataset

torch = lazy_load("torch")
torchaudio = lazy_load("torchaudio")
torchvision = lazy_load("torchvision")

if TYPE_CHECKING:
    import torch


class PyTorchDataset(DagsHubDataset, torch.utils.data.Dataset):
    def __init__(self, *args, **kwargs):
        self.tensorlib = Tensorizers
        super().__init__(*args, **kwargs)
        self.type = "torch"


class _BaseDataLoaderIter:
    def __next__(self) -> Any:
        return self.post_hook(super().__next__())


class _SingleProcessDataLoaderIter(_BaseDataLoaderIter, torch.utils.data.dataloader._SingleProcessDataLoaderIter):
    def __init__(self, *args, post_hook, **kwargs):
        self.post_hook = post_hook
        super().__init__(*args, **kwargs)


class _MultiProcessingDataLoaderIter(_BaseDataLoaderIter, torch.utils.data.dataloader._MultiProcessingDataLoaderIter):
    def __init__(self, *args, post_hook, **kwargs):
        self.post_hook = post_hook
        super().__init__(*args, **kwargs)


class PyTorchDataLoader(torch.utils.data.DataLoader):
    def __init__(self, *args, post_hook=lambda x: x, **kwargs):
        super().__init__(*args, **kwargs)
        self.post_hook = post_hook
        self.dataset.order = list(self.sampler)
        if self.dataset.strategy == "background":
            Process(target=self.dataset.pull).start()

    def _get_iterator(self) -> "_BaseDataLoaderIter":
        if self.num_workers == 0:
            return _SingleProcessDataLoaderIter(self, post_hook=self.post_hook)
        else:
            self.check_worker_number_rationality()
            return _MultiProcessingDataLoaderIter(self, post_hook=self.post_hook)


class Tensorizers:
    @staticmethod
    def image(filepath: str) -> torch.Tensor:
        return torchvision.io.read_image(filepath).type(torch.float)

    @staticmethod
    def audio(filepath: str) -> torch.Tensor:
        return torchaudio.load(filepath).type(torch.float)

    @staticmethod
    def video(filepath: str) -> torch.Tensor:
        return torchvision.io.read_video(filepath).type(torch.float)

    @staticmethod
    def numeric(num: Union[float, int]) -> torch.Tensor:
        return torch.tensor(num, dtype=torch.float)



import builtins
import importlib
import io
import logging
import os
import re
import subprocess
import sys
from configparser import ConfigParser
from functools import wraps, cached_property
from multiprocessing import AuthenticationError
from os import PathLike
from pathlib import Path, PurePosixPath
from typing import Optional, TypeVar, Union, Dict, Set, Tuple, List, Any, Callable
from urllib.parse import urlparse, ParseResult

import dacite
from httpx import Response
from tenacity import retry, retry_if_result, stop_after_attempt, wait_exponential, before_sleep_log, RetryError

from dagshub.common import config, is_inside_notebook, is_inside_colab
from dagshub.common.api.repo import RepoAPI, CommitNotFoundError
from dagshub.common.api.responses import ContentAPIEntry, StorageContentAPIResult
from dagshub.common.helpers import http_request, get_project_root, log_message
from dagshub.streaming.dataclasses import DagshubPath
from dagshub.streaming.errors import FilesystemAlreadyMountedError

# Pre 3.11 - need to patch _NormalAccessor for _pathlib, because it pre-caches open and other functions.
# In 3.11 _NormalAccessor was removed
PRE_PYTHON3_11 = sys.version_info.major == 3 and sys.version_info.minor < 11
if PRE_PYTHON3_11:
    from pathlib import _NormalAccessor as _pathlib  # noqa: E402


T = TypeVar("T")
logger = logging.getLogger(__name__)


def wrapreturn(wrappertype):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return wrappertype(func(*args, **kwargs))

        return wrapper

    return decorator


class dagshub_ScandirIterator:
    def __init__(self, iterator):
        self._iterator = iterator

    def __iter__(self):
        return self._iterator

    def __next__(self):
        return self._iterator.__next__()

    def __enter__(self):
        return self

    def __exit__(self, *args):
        return self


SPECIAL_FILE = Path(".dagshub-streaming")


def _is_server_error(resp: Response):
    return resp.status_code >= 500


# TODO: Singleton metaclass that lets us keep a "main" DvcFilesystem instance
class DagsHubFilesystem:
    """
    A DagsHub-repo aware filesystem class

    :param project_root: Path to the git repository with the repo.
        If None, traverse up the filesystem from the current dir until we find a git repo
    :param repo_url: URL to the DagsHub repository.
        If None, URL is received from the git configuration
    :param branch: Explicitly sets a branch/commit revision to work with
        If None, branch is received from the git configuration
    :param token: DagsHub API token
    :param username: DagsHub username (as an alternative to using the token)
    :param password: DagsHub password (as an alternative to using the token)
    :param timeout: Timeout in seconds for HTTP requests.
        Influences all requests except for file download, which has no timeout
    :param exclude_globs: One or more glob patterns to exclude from looking up on the server
        This is useful in case your framework tries to look up cached files on disk that might not be there.
        Example: YOLO and .npy files
    :param frameworks: List of frameworks that need custom patched openers.
        Right now the following is supported:

        - ``transformers`` - patches ``safetensors``
    """

    already_mounted_filesystems: Dict[Path, "DagsHubFilesystem"] = {}
    hooked_instance: Optional["DagsHubFilesystem"] = None

    # Framework-specific override functions.
    # These functions will be patched with a function that calls fs.open() before calling the original function
    # Classes are marked by $, so if you need to change a static/class method, use module.$class.func
    _framework_override_map: Dict[str, List[str]] = {
        "transformers": ["safetensors.safe_open", "tokenizers.$Tokenizer.from_file"],
    }

    def __init__(
        self,
        project_root: Optional["PathLike | str"] = None,
        repo_url: Optional[str] = None,
        branch: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        token: Optional[str] = None,
        timeout: Optional[int] = None,
        exclude_globs: Optional[Union[List[str], str]] = None,
        frameworks: Optional[List[str]] = None,
    ):
        # Find root directory of Git project
        if not project_root:
            try:
                self.project_root = get_project_root(Path(os.path.abspath(".")))
            except ValueError:
                raise ValueError(
                    "Could not find a git repo. Either run the function inside of a git repo, "
                    "specify `project_root` with the path to a cloned DagsHub repository, "
                    "or specify `repo_url` (url of repo on DagsHub) and "
                    "`project_root` (path to the folder where to mount the filesystem) arguments"
                )

        else:
            self.project_root = Path(os.path.abspath(project_root))
        del project_root
        # TODO: if no Git project found, search for .dvc project?

        if not repo_url:
            remotes = self.get_remotes_from_git_config()
            if len(remotes) > 0:
                repo_url = remotes[0]
            else:
                raise ValueError("No DagsHub git remote detected, please specify repo_url= argument or --repo_url flag")

        self.user_specified_branch = branch
        self.parsed_repo_url = urlparse(repo_url)
        # Key: path, value: dict of {name, type} on that path (in remote)
        self.remote_tree: Dict[str, Dict[str, str]] = {}
        self.frameworks = frameworks

        # Determine if any authentication is needed
        self.username = username or config.username
        self.password = password or config.password
        self.token = token or config.token
        self.timeout = timeout or config.http_timeout

        if exclude_globs is None:
            exclude_globs = []
        elif exclude_globs is str:
            exclude_globs = [exclude_globs]

        self.exclude_globs: List[str] = exclude_globs

        self._listdir_cache: Dict[str, Optional[Tuple[List[ContentAPIEntry], bool]]] = {}

        self._api = self._generate_repo_api(self.parsed_repo_url)

        self.check_project_root_use()

        # Check that the repo is accessible by accessing the content root
        response = self._api_listdir(DagshubPath(self, self.project_root, Path(), Path()))
        if response is None:
            # TODO: Check .dvc/config{,.local} for credentials
            raise AuthenticationError("DagsHub credentials required, however none provided or discovered")

        self._storages = self._api.get_connected_storages()

    def _generate_repo_api(self, repo_url: ParseResult) -> RepoAPI:
        host = f"{repo_url.scheme}://{repo_url.netloc}"
        repo = repo_url.path
        return RepoAPI(repo=repo, host=host, auth=self.auth)

    @cached_property
    def _current_revision(self) -> str:
        """
        Gets current revision on repo:
        - If User specified a branch, returns HEAD of that brunch on the remote
        - If branch wasn't detected, returns HEAD of default branch in the speficied remote.
        - If HEAD is a branch, tries to find a dagshub remote associated with it and get its HEAD
        - If HEAD is a commit revision, checks that the commit exists on DagsHub
        """

        if self.user_specified_branch:
            branch = self.user_specified_branch
        else:
            try:
                with open(self.project_root / ".git/HEAD") as head_file:
                    head = head_file.readline().strip()
                if head.startswith("ref"):
                    branch = head.split("/")[-1]
                else:
                    # contents of HEAD is the revision - check that this commit exists on remote
                    if self.is_commit_on_remote(head):
                        return head
                    else:
                        raise RuntimeError(
                            f"Current HEAD ({head}) doesn't exist on the remote. "
                            f"Please push your changes to the remote or checkout a tracked branch."
                        )

            except FileNotFoundError:
                logger.debug(
                    "Couldn't get branch info from local git repository, "
                    + "fetching default branch from the remote..."
                )
                branch = self._api.default_branch

        # check if it is a commit sha, in that case do not load the sha
        sha_regex = re.compile(r"^([a-f0-9]){5,40}$")
        if sha_regex.match(branch):
            try:
                self._api.get_commit_info(branch)
                return branch
            except CommitNotFoundError:
                pass

        return self._api.last_commit_sha(branch)

    def is_commit_on_remote(self, sha1):
        try:
            self._api.get_commit_info(sha1)
            return True
        except CommitNotFoundError:
            return False

    def check_project_root_use(self):
        """
        Checks that there's no other filesystem being mounted at the current project root
        If there is one, throw an error

        :meta private:
        """

        def is_subpath(a: Path, b: Path) -> bool:
            # Checks if either a or b are subpaths of each other
            a_str = a.as_posix()
            b_str = b.as_posix()
            return a_str.startswith(b_str) or b_str.startswith(a_str)

        for p, f in DagsHubFilesystem.already_mounted_filesystems.items():
            if is_subpath(p, self.project_root):
                raise FilesystemAlreadyMountedError(self.project_root, f.parsed_repo_url.path[1:], f._current_revision)

        DagsHubFilesystem.already_mounted_filesystems[self.project_root] = self

    def get_remote_branch_head(self, branch):
        """
        Get the head commit ID of a remote branch.

        Args:
            branch (str): The name of the remote branch.

        Raises:
            RuntimeError: Raised if there is an issue when trying to get the head of the branch.

        Returns:
            str: The commit ID of the head of the remote branch.

        :meta private:
        """
        url = self.get_api_url(f"/api/v1/repos{self.parsed_repo_url.path}/branches/{branch}")
        resp = self.http_get(url)
        if resp.status_code != 200:
            raise RuntimeError(
                f"Got status {resp.status_code} while trying to get head of branch {branch}. \r\n"
                f"Response body: {resp.content}"
            )
        return resp.json()["commit"]["id"]

    @property
    def auth(self):
        import dagshub.auth

        if self.username is not None and self.password is not None:
            return self.username, self.password

        try:
            return dagshub.auth.get_authenticator()
        except dagshub.auth.OauthNonInteractiveShellException:
            logger.debug("Failed to perform OAuth in a non interactive shell")

        # Try to fetch credentials from the git credential file
        proc = subprocess.run(["git", "credential", "fill"], input=f"url={self.repo_url}".encode(), capture_output=True)
        answer = {line[: line.index("=")]: line[line.index("=") + 1 :] for line in proc.stdout.decode().splitlines()}
        if "username" in answer and "password" in answer:
            return answer["username"], answer["password"]

    def get_remotes_from_git_config(self) -> List[str]:
        """
        Get the list of DagsHub remotes from the Git configuration.

        Returns:
            List[str]: A list of DagsHub remote URLs.

        :meta private:
        """
        # Get URLs of dagshub remotes
        git_config = ConfigParser()
        git_config.read(Path(self.project_root) / ".git/config")
        git_remotes = [urlparse(git_config[remote]["url"]) for remote in git_config if remote.startswith("remote ")]
        res_remotes = []
        for remote in git_remotes:
            if remote.hostname != config.hostname:
                continue
            remote = remote._replace(netloc=remote.hostname)
            remote = remote._replace(path=re.compile(r"(\.git)?/?$").sub("", remote.path))
            res_remotes.append(remote.geturl())
        return res_remotes

    def __del__(self):
        self.cleanup()

    def cleanup(self):
        # Remove from map of mounted filesystems
        if hasattr(self, "project_root") and self.project_root in DagsHubFilesystem.already_mounted_filesystems:
            DagsHubFilesystem.already_mounted_filesystems.pop(self.project_root)

    def _parse_path(self, file: Union[str, PathLike, int]) -> DagshubPath:
        orig_path = Path(file)
        if isinstance(file, int):
            return DagshubPath(self, None, None, orig_path)
        if file == "":
            return DagshubPath(self, None, None, orig_path)
        abspath = Path(os.path.abspath(file))
        try:
            relpath = abspath.relative_to(os.path.abspath(self.project_root))
            if str(relpath).startswith("<"):
                return DagshubPath(self, abspath, None, orig_path)
            return DagshubPath(self, abspath, relpath, orig_path)
        except ValueError:
            return DagshubPath(self, abspath, None, orig_path)

    def _special_file(self):
        # TODO Include more information in this file
        return b"v0\n"

    def open(self, file, mode="r", buffering=-1, encoding=None, errors=None, newline=None, closefd=True, opener=None):
        """
        NOTE: This is a wrapper function for python's built-in file operations
            (https://docs.python.org/3/library/functions.html#open)

        Open a file for reading or writing, with support for special files and DagsHub integration.

        Args:
            file (Union[str, int, bytes]): The file to be opened.
                It can be a path (str), file descriptor (int), or bytes-like object.
            mode (str, optional): The mode in which the file should be opened. Defaults to 'r'.
            buffering (int, optional): The buffering value. Defaults to -1.
            encoding (str, optional): The encoding to use when reading the file. Defaults to None.
            errors (str, optional): The error handling strategy. Defaults to None.
            newline (str, optional): The newline parameter. Defaults to None.
            closefd (bool, optional): Whether to close the file descriptor. Defaults to True.
            opener (callable, optional): The file opener. Defaults to None.

        Returns:
            File object: A file object representing the opened file.

        :meta private:
        """
        # FD passthrough
        if type(file) is int:
            return self.__open(file, mode, buffering, encoding, errors, newline, closefd)

        if type(file) is bytes:
            file = os.fsdecode(file)
        path = self._parse_path(file)
        if path.is_in_repo:
            if opener is not None:
                raise NotImplementedError("DagsHub's patched open() does not support custom openers")
            if path.is_passthrough_path:
                return self.__open(path.absolute_path, mode, buffering, encoding, errors, newline, closefd)
            elif path.relative_path == SPECIAL_FILE:
                return io.BytesIO(self._special_file())
            else:
                try:
                    return self.__open(path.absolute_path, mode, buffering, encoding, errors, newline, closefd)
                except FileNotFoundError as err:
                    # Open for reading - try to download the file
                    if "r" in mode:
                        try:
                            resp = self._api_download_file_git(path)
                        except RetryError:
                            raise RuntimeError(f"Couldn't download {path.relative_path} after multiple attempts")
                        if resp.status_code < 400:
                            self._mkdirs(path.absolute_path.parent)
                            # TODO: Handle symlinks
                            with self.__open(path.absolute_path, "wb") as output:
                                output.write(resp.content)
                            return self.__open(path.absolute_path, mode, buffering, encoding, errors, newline, closefd)
                        elif resp.status_code == 404:
                            raise FileNotFoundError(f"Error finding {path.relative_path} in repo or on DagsHub")
                        else:
                            raise RuntimeError(
                                f"Got response code {resp.status_code} from DagsHub while downloading file"
                                f" {path.relative_path}"
                            )
                    # Write modes - make sure that the folder is a tracked folder (create if doesn't exist on disk),
                    # and then let the user write to file
                    else:
                        try:
                            # Using the fact that stat creates tracked dirs (but still throws on nonexistent dirs)
                            _ = self.stat(path.absolute_path.parent)
                        except FileNotFoundError:
                            raise err
                        # Try to download the file if we're in append modes
                        if "a" in mode or "+" in mode:
                            try:
                                resp = self._api_download_file_git(path)
                            except RetryError:
                                raise RuntimeError(f"Couldn't download {path.relative_path} after multiple attempts")
                            if resp.status_code < 400:
                                with self.__open(path.absolute_path, "wb") as output:
                                    output.write(resp.content)
                        return self.__open(path.absolute_path, mode, buffering, encoding, errors, newline, closefd)

        else:
            return self.__open(file, mode, buffering, encoding, errors, newline, closefd, opener)

    def os_open(self, path, flags, mode=0o777, *, dir_fd=None):
        """
        os.open is supposed to be lower level, but it's still being used by e.g. Pathlib
        We're trying to wrap around it here, by parsing flags and calling the higher-level open
        Caveats: list of flags being handled is not exhaustive + mode doesn't work
                 (because we lose them when passing to builtin open)
        WARNING: DO NOT patch actual os.open with it, because the builtin uses os.open.
                 This is only for the purposes of patching pathlib.open in Python 3.9 and below.
                 Since Python 3.10 pathlib uses io.open, and in Python 3.11 they removed the accessor completely

        :meta private:
        """
        if dir_fd is not None:  # If dir_fd supplied, path is relative to that dir's fd, will handle in the future
            logger.debug("fs.os_open - NotImplemented")
            raise NotImplementedError("DagsHub's patched os.open() (for pathlib only) does not support dir_fd")
        path = self._parse_path(path)
        if path.is_in_repo:
            try:
                open_mode = "r"
                # Write modes - calling in append mode,
                # This way we create the intermediate folders if file doesn't exist, but the folder it's in does
                # Append so we don't truncate the file
                if not (flags & os.O_RDONLY):
                    open_mode = "a"
                logger.debug("fs.os_open - trying to materialize path")
                self.open(path.absolute_path, mode=open_mode).close()
                logger.debug("fs.os_open - successfully materialized path")
            except FileNotFoundError:
                logger.debug("fs.os_open - failed to materialize path, os.open will throw")
        return os.open(path.absolute_path, flags, mode, dir_fd=dir_fd)

    def stat(self, path, *args, dir_fd=None, follow_symlinks=True):
        """
        NOTE: This is a wrapper function for python's built-in file operations
            (https://docs.python.org/3/library/os.html#os.stat)

        Get the status of a file or directory, including support for special files and DagsHub integration.

        Args:
            path (Union[str, int, bytes]): The path of the file or directory to get the status for.
                It can be a path (str), file descriptor (int), or bytes-like object.
            dir_fd (int, optional): File descriptor of the directory. Defaults to None.
            follow_symlinks (bool, optional): Whether to follow symbolic links. Defaults to True.

        Returns:
            collections.namedtuple: A namedtuple containing the file status information.

        :meta private:
        """
        # FD passthrough
        if type(path) is int:
            return self.__stat(path, *args, dir_fd=dir_fd, follow_symlinks=follow_symlinks)

        if type(path) is bytes:
            path = os.fsdecode(path)
        if dir_fd is not None or not follow_symlinks:
            logger.debug("fs.stat - NotImplemented")
            raise NotImplementedError("DagsHub's patched stat() does not support dir_fd or follow_symlinks")
        parsed_path = self._parse_path(path)
        # todo: remove False
        if parsed_path.is_in_repo:
            logger.debug("fs.stat - is relative path")
            if parsed_path.is_passthrough_path:
                return self.__stat(parsed_path.absolute_path)
            elif parsed_path.relative_path == SPECIAL_FILE:
                return dagshub_stat_result(self, path, is_directory=False, custom_size=len(self._special_file()))
            else:
                try:
                    logger.debug(f"fs.stat - calling __stat - relative_path: {path}")
                    return self.__stat(parsed_path.absolute_path)
                except FileNotFoundError as err:
                    logger.debug("fs.stat - FileNotFoundError")
                    logger.debug(f"remote_tree: {self.remote_tree}")
                    parent_path = parsed_path.relative_path.parent
                    if str(parent_path) not in self.remote_tree:
                        try:
                            # Run listdir to update cache
                            self.listdir(self.project_root / parent_path)
                        except FileNotFoundError:
                            raise err

                    cached_remote_parent_tree = self.remote_tree.get(str(parent_path))
                    logger.debug(f"cached_remote_parent_tree: {cached_remote_parent_tree}")

                    if cached_remote_parent_tree is None:
                        raise err

                    filetype = cached_remote_parent_tree.get(parsed_path.name)
                    if filetype is None:
                        raise err

                    if filetype == "file":
                        return dagshub_stat_result(self, path, is_directory=False)
                    elif filetype == "dir":
                        self._mkdirs(parsed_path.absolute_path)
                        return self.__stat(parsed_path.absolute_path)
                    else:
                        raise RuntimeError(f"Unknown file type {filetype} for path {str(parsed_path)}")

        else:
            return self.__stat(path, follow_symlinks=follow_symlinks)

    def chdir(self, path):
        """
         NOTE: This is a wrapper function for python's built-in file operations
            (https://docs.python.org/3/library/os.html#os.chdir)

        Change the current working directory to the specified path, with support for DagsHub integration.

        Args:
            path (Union[str, int, bytes]): The path to change the current working directory to.
                It can be a path (str), file descriptor (int), or bytes-like object.

        :meta private:
        """
        # FD check
        if type(path) is int:
            return self.__chdir(path)

        if type(path) is bytes:
            path = os.fsdecode(path)
        parsed_path = self._parse_path(path)
        if parsed_path.is_in_repo:
            try:
                self.__chdir(parsed_path.absolute_path)
            except FileNotFoundError:
                resp = self._api_listdir(parsed_path)
                # FIXME: if path is file, return FileNotFound instead of the listdir error
                if resp is not None:
                    self._mkdirs(parsed_path.absolute_path)
                    self.__chdir(parsed_path.absolute_path)
                else:
                    raise
        else:
            self.__chdir(path)

    def listdir(self, path="."):
        """
        NOTE: This is a wrapper function for python's built-in file operations
            (https://docs.python.org/3/library/os.html#os.listdir)

        List the contents of a directory, including support for DagsHub integration.

        Args:
            path (str, optional): The path of the directory to list. Defaults to '.'.

        Raises:
            error: If an error occurs during the operation, it will be raised.

        Returns: A list of directory contents. If 'path' is a bytes object, the results will also be in bytes.

        :meta private:
        """
        # FD check
        if type(path) is int:
            return self.__listdir(path)

        # listdir needs to return results for bytes path arg also in bytes
        is_bytes_path_arg = type(path) is bytes

        def encode_results(res):
            res = list(res)
            if is_bytes_path_arg:
                res = [os.fsencode(p) for p in res]
            return res

        if is_bytes_path_arg:
            str_path = os.fsdecode(path)
        else:
            str_path = path
        parsed_path = self._parse_path(str_path)
        if parsed_path.is_in_repo:
            if parsed_path.is_passthrough_path:
                return self.__listdir(parsed_path.original_path)
            else:
                dircontents: Set[str] = set()
                error = None
                try:
                    dircontents.update(self.__listdir(parsed_path.original_path))
                except FileNotFoundError as e:
                    error = e
                dircontents.update(
                    special.name
                    for special in self._get_special_paths(
                        parsed_path, self.project_root_dagshub_path, is_bytes_path_arg
                    )
                )
                # If we're accessing .dagshub/storage/s3/ we don't need to access the API, return straight away
                len_parts = len(parsed_path.relative_path.parts)
                if 0 < len_parts <= 3 and parsed_path.relative_path.parts[0] == ".dagshub":
                    return encode_results(dircontents)
                resp = self._api_listdir(parsed_path)
                if resp is not None:
                    dircontents.update(Path(f.path).name for f in resp)
                    self.remote_tree[str(parsed_path.relative_path)] = {Path(f.path).name: f.type for f in resp}
                    return encode_results(dircontents)
                else:
                    if error is not None:
                        raise error
                    else:
                        return encode_results(dircontents)

        else:
            return self.__listdir(path)

    @cached_property
    def project_root_dagshub_path(self):
        return DagshubPath(absolute_path=self.project_root, relative_path=Path(), original_path=Path(), fs=self)

    @wrapreturn(dagshub_ScandirIterator)
    def scandir(self, path="."):
        # FD check
        if type(path) is int:
            for direntry in self.__scandir(path):
                yield direntry
            return
        # scandir needs to return name and path as bytes, if entry arg is bytes
        is_bytes_path_arg = type(path) is bytes
        if is_bytes_path_arg:
            str_path = os.fsdecode(path)
        else:
            str_path = path
        parsed_path = self._parse_path(str_path)
        if parsed_path.is_in_repo and not parsed_path.is_passthrough_path:
            path = Path(str_path)
            local_filenames = set()
            try:
                for direntry in self.__scandir(path):
                    local_filenames.add(direntry.name)
                    yield direntry
            except FileNotFoundError:
                pass
            for special_entry in self._get_special_paths(
                parsed_path, self.project_root_dagshub_path / path, is_bytes_path_arg
            ):
                if special_entry.path not in local_filenames:
                    yield special_entry
            # Mix in the results from the API
            resp = self._api_listdir(parsed_path)
            if resp is not None:
                for f in resp:
                    name = PurePosixPath(f.path).name
                    if name not in local_filenames:
                        yield dagshub_DirEntry(self, parsed_path / name, f.type == "dir", is_binary=is_bytes_path_arg)
        else:
            for entry in self.__scandir(path):
                yield entry

    def _get_special_paths(
        self, dh_path: DagshubPath, relative_to: DagshubPath, is_binary: bool
    ) -> Set["dagshub_DirEntry"]:
        def generate_entry(path, is_directory):
            if isinstance(path, str):
                path = Path(path)
            return dagshub_DirEntry(self, relative_to / path, is_directory=is_directory, is_binary=is_binary)

        has_storages = len(self._storages) > 0
        res = set()
        str_path = dh_path.relative_path.as_posix()
        if str_path == ".":
            res.add(generate_entry(SPECIAL_FILE, False))
            if has_storages:
                res.add(generate_entry(".dagshub", True))
        elif str_path.startswith(".dagshub") and has_storages:
            storage_paths = [s.path_in_mount for s in self._storages]
            for sp in storage_paths:
                try:
                    relpath = sp.relative_to(dh_path.relative_path)
                    if relpath != Path():
                        res.add(generate_entry(relpath.parts[0], True))
                except ValueError:
                    continue
        return res

    def _api_listdir(self, path: DagshubPath, include_size: bool = False) -> Optional[List[ContentAPIEntry]]:
        response, hit = self._check_listdir_cache(path.relative_path.as_posix(), include_size)
        if hit:
            return response
        params: Dict[str, Any] = {"include_size": "true"} if include_size else {}
        if path.is_storage_path:
            params["paging"] = True
        url = self._content_url_for_path(path)

        def _get() -> Optional[Response]:
            resp = self.http_get(url, params=params, headers=config.requests_headers)
            if resp.status_code == 404:
                logger.debug(f"Got HTTP code {resp.status_code} while listing {path}, no results will be returned")
                return None
            elif resp.status_code >= 400:
                logger.warning(f"Got HTTP code {resp.status_code} while listing {path}, no results will be returned")
                return None
            return resp

        response = _get()
        if response is None:
            return None
        res: List[ContentAPIEntry] = []
        # Storage - token pagination, different return structure + if there's a token we do another request
        if path.is_storage_path:
            result = dacite.from_dict(StorageContentAPIResult, response.json())
            res += result.entries
            while result.next_token is not None:
                params["from_token"] = result.next_token
                new_resp = _get()
                if new_resp is None:
                    return None
                result = dacite.from_dict(StorageContentAPIResult, new_resp.json())
                res += result.entries
        else:
            for entry_raw in response.json():
                entry = dacite.from_dict(ContentAPIEntry, entry_raw)
                # Ignore storage root entries, we handle them separately in a different place
                if entry.type == "storage":
                    continue
                res.append(entry)

        self._listdir_cache[path.relative_path.as_posix()] = (res, include_size)
        return res

    def _check_listdir_cache(self, path: str, include_size: bool) -> Tuple[Optional[List[ContentAPIEntry]], bool]:
        # Checks that path has a pre-cached response
        # If include_size is True, but only a response without size is cached, that's a cache miss
        if path in self._listdir_cache:
            cache_val, with_size = self._listdir_cache[path]
            if not include_size or (include_size and with_size):
                return cache_val, True
        return None, False

    def _content_url_for_path(self, path: DagshubPath):
        if not path.is_in_repo:
            raise RuntimeError(f"Can't access path {path.absolute_path} outside of repo")
        str_path = path.relative_path.as_posix()
        if path.is_storage_path:
            path_to_access = str_path[len(".dagshub/storage/") :]
            return self._api.storage_content_api_url(path_to_access)
        return self._api.content_api_url(str_path, self._current_revision)

    def _raw_url_for_path(self, path: DagshubPath):
        if not path.is_in_repo:
            raise RuntimeError(f"Can't access path {path.absolute_path} outside of repo")
        str_path = path.relative_path.as_posix()
        if path.is_storage_path:
            path_to_access = str_path[len(".dagshub/storage/") :]
            return self._api.storage_raw_api_url(path_to_access)
        return self._api.raw_api_url(str_path, self._current_revision)

    @retry(
        retry=retry_if_result(_is_server_error),
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=4, max=10),
        before_sleep=before_sleep_log(logger, logging.WARNING),
    )
    def _api_download_file_git(self, path: DagshubPath):
        resp = self.http_get(self._raw_url_for_path(path), headers=config.requests_headers, timeout=None)
        return resp

    def http_get(self, path: str, **kwargs):
        timeout = self.timeout
        if "timeout" in kwargs:
            timeout = kwargs["timeout"]
            del kwargs["timeout"]
        return http_request("GET", path, auth=self.auth, timeout=timeout, **kwargs)

    def install_hooks(self):
        """
        Install hooks to override default file and directory operations with DagsHub-aware functionality.

        This method patches the standard Python I/O operations such as ``open``,
        ``stat``, ``listdir``, ``scandir``, and ``chdir`` with DagsHub-aware equivalents.
        Works inside a notebook and with Pathlib.

        If ``install_hooks()`` have already been called before, this method does nothing.

        Example::

            dagshub_fs = DagsHubFilesystem()
            dagshub_fs.install_hooks()

            with open("src/file_in_repo.txt") as f:
                print(f.read())

        Call :func:`~DagsHubFilesystem.uninstall_hooks` to undo the monkey patching.
        """
        if not hasattr(self.__class__, f"_{self.__class__.__name__}__unpatched"):
            # TODO: DRY this dictionary. i.e. __open() links cls.__open
            #  and io.open even though this dictionary links them
            #  Cannot use a dict as the source of truth because type hints rely on
            #  __get_unpatched inferring the right type
            self.__class__.__unpatched = {
                "open": builtins.open,
                "stat": os.stat,
                "listdir": os.listdir,
                "scandir": os.scandir,
                "chdir": os.chdir,
            }
            if PRE_PYTHON3_11:
                self.__class__.__unpatched["pathlib_open"] = _pathlib.open

        # IPython patches io.open to its own override, so we need to overwrite that also
        # More at _modified_open function in IPython sources:
        # https://github.com/ipython/ipython/blob/main/IPython/core/interactiveshell.py
        if is_inside_notebook() and not is_inside_colab():
            import IPython.core.interactiveshell

            instance = IPython.core.interactiveshell.InteractiveShell._instance  # noqa
            if instance is not None and hasattr(instance, "user_ns") and "open" in instance.user_ns:
                self.__class__.__unpatched["notebook_open"] = instance.user_ns["open"]
                instance.user_ns["open"] = self.open

        io.open = builtins.open = self.open
        os.stat = self.stat
        os.listdir = self.listdir
        os.scandir = self.scandir
        os.chdir = self.chdir
        if PRE_PYTHON3_11:
            if sys.version_info.minor == 10:
                # Python 3.10 - pathlib uses io.open
                _pathlib.open = self.open
            else:
                # Python <=3.9 - pathlib uses os.open
                _pathlib.open = self.os_open
            _pathlib.stat = self.stat
            _pathlib.listdir = self.listdir
            _pathlib.scandir = self.scandir

        self._install_framework_hooks()

        DagsHubFilesystem.hooked_instance = self

        msg = (
            f'Repository "{self._api.full_name}" is now hooked at path "{self.project_root}".\n'
            f"Any calls to Python file access function like open() and listdir() inside "
            f"of this directory will include results from the repository."
        )
        log_message(msg, logger)

    _framework_key_prefix = "framework_"

    def _install_framework_hooks(self):
        """
        Installs custom hook functions for frameworks
        """
        if self.frameworks is None:
            return
        for framework in self.frameworks:
            if framework not in self._framework_override_map:
                logger.warning(f"Framework {framework} not available for override, skipping")
                continue
            funcs = self._framework_override_map[framework]
            for func in funcs:
                module_name, func_name = func.rsplit(".", 1)
                class_name = None
                patch_class = None

                # Handle static class methods - we'll need to get the class from the module first
                if "$" in module_name:
                    module_name, class_name = module_name.split("$")
                    # Get rid of the . in the module name
                    module_name = module_name[:-1]

                try:
                    patch_module = importlib.import_module(module_name)
                    if class_name is not None:
                        patch_class = getattr(patch_module, class_name)
                        orig_fn = getattr(patch_class, func_name)
                    else:
                        orig_fn = getattr(patch_module, func_name)
                except ModuleNotFoundError:
                    logger.warning(f"Module [{module_name}] not found, so function [{func}] isn't being patched")
                    continue
                except AttributeError:
                    logger.warning(f"Function [{func}] not found, not patching it")
                    continue
                self.__class__.__unpatched[f"{self._framework_key_prefix}{func}"] = orig_fn
                if patch_class is not None:
                    setattr(patch_class, func_name, self._passthrough_decorator(orig_fn))
                else:
                    setattr(patch_module, func_name, self._passthrough_decorator(orig_fn))

    def _passthrough_decorator(self, orig_func, filearg: Union[int, str] = 0) -> Callable:
        """
        Decorator function over some other random function that assumes a file exists locally,
        but isn't using python's open(). These might be C++/Rust functions that use their respective opens.
        Examples: opencv, anything using pyo3

        Working around the problem by first calling open().close() to get the file.

        :param orig_func: the original function that needs to be called
        :param filearg: int or string, which arg/kwarg to use to get the filename
        :return: Wrapped orig_func
        """

        def passed_through(*args, **kwargs):
            if type(filearg) is str:
                filename = kwargs[filearg]
            else:
                filename = args[filearg]
            self.open(filename).close()
            return orig_func(*args, **kwargs)

        return passed_through

    @classmethod
    def uninstall_hooks(cls):
        """
        Reverses the changes made by :func:`install_hooks`, bringing back the builtin file I/O functions.
        """
        if hasattr(cls, f"_{cls.__name__}__unpatched"):
            io.open = builtins.open = cls.__unpatched["open"]
            os.stat = cls.__unpatched["stat"]
            os.listdir = cls.__unpatched["listdir"]
            os.scandir = cls.__unpatched["scandir"]
            os.chdir = cls.__unpatched["chdir"]
            if PRE_PYTHON3_11:
                _pathlib.open = cls.__unpatched["pathlib_open"]
                _pathlib.stat = cls.__unpatched["stat"]
                _pathlib.listdir = cls.__unpatched["listdir"]
                _pathlib.scandir = cls.__unpatched["scandir"]

            if "notebook_open" in cls.__unpatched:
                import IPython.core.interactiveshell

                instance = IPython.core.interactiveshell.InteractiveShell._instance  # noqa
                if instance is not None and hasattr(instance, "user_ns"):
                    instance.user_ns["open"] = cls.__unpatched["notebook_open"]

            cls._uninstall_framework_hooks()

        if DagsHubFilesystem.hooked_instance is not None:
            DagsHubFilesystem.hooked_instance.cleanup()
            DagsHubFilesystem.hooked_instance = None

    @classmethod
    def _uninstall_framework_hooks(cls):
        for func in list(filter(lambda key: key.startswith(cls._framework_key_prefix), cls.__unpatched.keys())):
            orig_fn = cls.__unpatched[func]
            orig_func_name = func

            func = func[len(cls._framework_key_prefix) :]
            module_name, func_name = func.rsplit(".", 1)
            class_name = None

            if "$" in module_name:
                module_name, class_name = module_name.split("$")
                # Get rid of the . in the module name
                module_name = module_name[:-1]

            m = importlib.import_module(module_name)
            if class_name is not None:
                patch_class = getattr(m, class_name)
                setattr(patch_class, func_name, orig_fn)
            else:
                setattr(m, func_name, orig_fn)

            del cls.__unpatched[orig_func_name]

    def _mkdirs(self, absolute_path: Path):
        for parent in list(absolute_path.parents)[::-1]:
            try:
                self.__stat(parent)
            except (OSError, ValueError):
                os.mkdir(parent)
        try:
            self.__stat(absolute_path)
        except (OSError, ValueError):
            os.mkdir(absolute_path)

    @classmethod
    def __get_unpatched(cls, key, alt: T) -> T:
        if hasattr(cls, f"_{cls.__name__}__unpatched"):
            return cls.__unpatched[key]
        else:
            return alt

    @property
    def __open(self):
        return self.__get_unpatched("open", builtins.open)

    @property
    def __stat(self):
        return self.__get_unpatched("stat", os.stat)

    @property
    def __listdir(self):
        return self.__get_unpatched("listdir", os.listdir)

    @property
    def __scandir(self):
        return self.__get_unpatched("scandir", os.scandir)

    @property
    def __chdir(self):
        return self.__get_unpatched("chdir", os.chdir)


def install_hooks(
    project_root: Optional[PathLike] = None,
    repo_url: Optional[str] = None,
    branch: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    token: Optional[str] = None,
    timeout: Optional[int] = None,
    exclude_globs: Optional[Union[List[str], str]] = None,
    frameworks: Optional[List[str]] = None,
):
    """
    Monkey patches builtin Python functions to make them DagsHub-repo aware.
    Patched functions are: ``open()``, ``os.listdir()``, ``os.scandir()``, ``os.stat()`` and \
    pathlib's functions that use them

    Calling this function is equivalent to creating a :class:`DagsHubFilesystem` object
    and calling its :func:`install_hooks() <dagshub.streaming.DagsHubFilesystem.install_hooks>` method

    For argument documentation, read :class:`.DagsHubFilesystem`

    Call :func:`uninstall_hooks` to undo the monkey patching.
    """
    fs = DagsHubFilesystem(
        project_root=project_root,
        repo_url=repo_url,
        branch=branch,
        username=username,
        password=password,
        token=token,
        timeout=timeout,
        exclude_globs=exclude_globs,
        frameworks=frameworks,
    )
    fs.install_hooks()


def uninstall_hooks():
    """
    Reverses the changes made by :func:`install_hooks`
    """
    DagsHubFilesystem.uninstall_hooks()


class dagshub_stat_result:
    def __init__(self, fs: "DagsHubFilesystem", path: DagshubPath, is_directory: bool, custom_size: int = None):
        self._fs = fs
        self._path = path
        self._is_directory = is_directory
        self._custom_size = custom_size
        assert not self._is_directory  # TODO make folder stats lazy?

    def __getattr__(self, name: str):
        if not name.startswith("st_"):
            raise AttributeError
        if hasattr(self, "_true_stat"):
            return os.stat_result.__getattribute__(self._true_stat, name)
        if name == "st_uid":
            return os.getuid()
        elif name == "st_gid":
            return os.getgid()
        elif name == "st_atime" or name == "st_mtime" or name == "st_ctime":
            return 0
        elif name == "st_mode":
            return 0o100644
        elif name == "st_size":
            if self._custom_size:
                return self._custom_size
            return 1100  # hardcoded size because size requests take a disproportionate amount of time
        self._fs.open(self._path)
        self._true_stat = self._fs._DagsHubFilesystem__stat(self._path.absolute_path)
        return os.stat_result.__getattribute__(self._true_stat, name)

    def __repr__(self):
        inner = repr(self._true_stat) if hasattr(self, "_true_stat") else "pending..."
        return f"dagshub_stat_result({inner}, path={self._path})"


class dagshub_DirEntry:
    def __init__(self, fs: "DagsHubFilesystem", path: DagshubPath, is_directory: bool = False, is_binary: bool = False):
        self._fs = fs
        self._path = path
        self._is_directory = is_directory
        self._is_binary = is_binary

    @property
    def name(self):
        # TODO: create decorator for delegation
        if hasattr(self, "_true_direntry"):
            name = self._true_direntry.name
        else:
            name = self._path.name
        return os.fsencode(name) if self._is_binary else name

    @property
    def path(self):
        if hasattr(self, "_true_direntry"):
            path = self._true_direntry.path
        else:
            path = str(self._path.original_path)
        return os.fsencode(path) if self._is_binary else path

    def is_dir(self):
        if hasattr(self, "_true_direntry"):
            return self._true_direntry.is_dir()
        else:
            return self._is_directory

    def is_file(self):
        if hasattr(self, "_true_direntry"):
            return self._true_direntry.is_file()
        else:
            # TODO: Symlinks should return false
            return not self._is_directory

    def stat(self):
        if hasattr(self, "_true_direntry"):
            return self._true_direntry.stat()
        else:
            return self._fs.stat(self._path.original_path)

    def __getattr__(self, name: str):
        if name == "_true_direntry":
            raise AttributeError
        if hasattr(self, "_true_direntry"):
            return os.DirEntry.__getattribute__(self._true_direntry, name)

        # Either create a dir, or download the file
        if self._is_directory:
            self._fs._mkdirs(self._path.absolute_path)
        else:
            self._fs.open(self._path.absolute_path)

        for direntry in self._fs._DagsHubFilesystem__scandir(self._path.original_path):
            if direntry.name == self._path.name:
                self._true_direntry = direntry
                return os.DirEntry.__getattribute__(self._true_direntry, name)
        else:
            raise FileNotFoundError

    def __repr__(self):
        cached = " (cached)" if hasattr(self, "_true_direntry") else ""
        return f"<dagshub_DirEntry '{self.name}'{cached}>"


# Used for testing purposes only
if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    fs = DagsHubFilesystem()
    fs.install_hooks()

__all__ = [DagsHubFilesystem.__name__, install_hooks.__name__]
